@@ -48,6 +48,10 @@ public:
                 }
             });
         }
+
+        for (auto p: m_pyobj_needs_decref) {
+            decref(p);
+        }
     }
 
     DeserializationBuffer(const DeserializationBuffer&) = delete;
@@ -240,7 +244,6 @@ public:
         while (true) {
             auto fieldAndWiretype = readFieldNumberAndWireType();
 
-
             if (fieldAndWiretype.second == WireType::END_COMPOUND) {
                 return;
             }
@@ -357,6 +360,12 @@ public:
         if (!ptr) {
             throw std::runtime_error("Corrupt data: can't write a null cache pointer.");
         }
+        if (which < 0) {
+            throw std::runtime_error("Invalid memo. Shouldn't be negative.");
+        }
+        if (which > m_cachedPointers.size() + 1000000) {
+            throw std::runtime_error("Suspicious memo is out of bounds.");
+        }
         while (which >= m_cachedPointers.size()) {
             m_cachedPointers.push_back(nullptr);
         }
@@ -372,6 +381,29 @@ public:
         return ptr;
     }
 
+    PyObject* addCachedPyObj(int32_t which, PyObject* ptr) {
+        if (!ptr) {
+            throw std::runtime_error("Corrupt data: can't write a null cache pointer.");
+        }
+        if (which < 0) {
+            throw std::runtime_error("Invalid memo. Shouldn't be negative.");
+        }
+        if (which > m_cachedPointers.size() + 1000000) {
+            throw std::runtime_error("Suspicious memo is out of bounds.");
+        }
+        while (which >= m_cachedPointers.size()) {
+            m_cachedPointers.push_back(nullptr);
+        }
+        if (m_cachedPointers[which]) {
+            throw std::runtime_error("Corrupt data: tried to write a recursive object multiple times");
+        }
+        m_cachedPointers[which] = (void*)ptr;
+
+        m_pyobj_needs_decref.push_back(ptr);
+
+        return ptr;
+    }
+
     void skipNextEncodedValue() {
         throw std::runtime_error("not implemented yet");
     }
@@ -544,4 +576,6 @@ private:
     //object (e.g. PyObject, Dict, etc). We pass a pointer to this
     //as the actual instance to the 'destroy' operation
     std::map<Type*, std::vector<void*> > m_needs_decref;
-};
\ No newline at end of file
+
+    std::vector<PyObject*> m_pyobj_needs_decref;
+};
