@@ -23,20 +23,22 @@
 #include <unordered_map>
 
 class HeldClass;
+class RefTo;
 class Function;
 class Class;
 
 //this takes an instance_ptr for a Class object (not the HeldClass)
 typedef void (*destructor_fun_type)(void* inst);
 
-typedef Function* function_signature_type;
+// represents a concrete call signature with positional arguments packed
+// into the tuple and the named arguments packed into the named tuple.
+// we'll end up with a each distinct function call signat
+typedef std::tuple<Type*, Tuple*, NamedTuple*> function_call_signature_type;
 
-typedef std::pair<std::string, Function*> method_signature_type;
+typedef std::pair<std::string, function_call_signature_type> method_call_signature_type;
 
 typedef void* untyped_function_ptr;
 
-void destroyClassInstance(instance_ptr classInstDestroy);
-
 class ConstCharPtrsAreEqual {
 public:
     bool operator()(const char* x, const char* y) const {
@@ -83,8 +85,8 @@ public:
         // these members are explicitly leaked so that the layout of the C class is
         // comprehensible to the llvm code layer. maps and sets have a nontrivial layout,
         // and so it's more stable to just hold them as pointers.
-        mDispatchIndices(*new std::map<method_signature_type, size_t>()),
-        mDispatchDefinitions(*new std::map<size_t, method_signature_type>()),
+        mDispatchIndices(*new std::map<method_call_signature_type, size_t>()),
+        mDispatchDefinitions(*new std::map<size_t, method_call_signature_type>()),
         mIndicesNeedingDefinition(*new std::set<size_t>())
     {
     }
@@ -111,18 +113,18 @@ public:
         allocateUpcastDispatchTables();
     }
 
-    size_t allocateMethodDispatch(std::string funcName, const function_signature_type& signature) {
+    size_t allocateMethodDispatch(std::string funcName, const function_call_signature_type& signature) {
         assertHoldingTheGil();
 
-        auto it = mDispatchIndices.find(method_signature_type(funcName, signature));
+        auto it = mDispatchIndices.find(method_call_signature_type(funcName, signature));
         if (it != mDispatchIndices.end()) {
             return it->second;
         }
 
         size_t newIndex = mDispatchIndices.size();
 
-        mDispatchIndices[method_signature_type(funcName, signature)] = newIndex;
-        mDispatchDefinitions[newIndex] = method_signature_type(funcName, signature);
+        mDispatchIndices[method_call_signature_type(funcName, signature)] = newIndex;
+        mDispatchDefinitions[newIndex] = method_call_signature_type(funcName, signature);
 
         // check if we need to allocate a bigger function pointer table. If we do,
         // we must leave the existing one in place, because compiled code may be
@@ -160,6 +162,10 @@ public:
             throw std::runtime_error("Tried to define a function pointer in a VTable.ClassDispatchTable to be null.");
         }
 
+        if (mFuncPtrs[index] == ptr) {
+            return;
+        }
+
         auto it = mIndicesNeedingDefinition.find(index);
         if (it == mIndicesNeedingDefinition.end()) {
             throw std::runtime_error("Tried to define a function pointer in a VTable.ClassDispatchTable twice.");
@@ -186,16 +192,20 @@ public:
         return mInterfaceClass;
     }
 
-    method_signature_type dispatchDefinitionForSlot(size_t slotIx) const {
+    method_call_signature_type dispatchDefinitionForSlot(size_t slotIx) const {
         auto it = mDispatchDefinitions.find(slotIx);
         if (it == mDispatchDefinitions.end()) {
-            throw std::runtime_error("Invalid slot");
+            throw std::runtime_error("Invalid slot " + format(slotIx));
         }
         return it->second;
     }
 
     void allocateUpcastDispatchTables();
 
+    untyped_function_ptr get(size_t slot) {
+        return mFuncPtrs[slot];
+    }
+
 private:
     // the class actually represented by this instance
     HeldClass* mImplementingClass;
@@ -217,9 +227,9 @@ private:
 
     size_t mFuncPtrsUsed;
 
-    std::map<method_signature_type, size_t>& mDispatchIndices;
+    std::map<method_call_signature_type, size_t>& mDispatchIndices;
 
-    std::map<size_t, method_signature_type>& mDispatchDefinitions;
+    std::map<size_t, method_call_signature_type>& mDispatchDefinitions;
 
     std::set<size_t>& mIndicesNeedingDefinition;
 };
@@ -228,7 +238,7 @@ class VTable {
 public:
     VTable(HeldClass* inClass) :
         mType(inClass),
-        mCompiledDestructorFun((destructor_fun_type)destroyClassInstance),
+        mCompiledDestructorFun(nullptr),
         mDispatchTables(nullptr)
     {
     }
@@ -239,7 +249,7 @@ public:
     base classes we might masquerade as.
     *****/
 
-    void finalize(ClassDispatchTable* dispatchers, long inInitializationBitByteCount) {
+    void finalize(ClassDispatchTable* dispatchers, long inInitializationBitByteCount, long count) {
         mDispatchTables = dispatchers;
         mInitializationBitByteCount = inInitializationBitByteCount;
     }
@@ -248,7 +258,7 @@ public:
         if (mCompiledDestructorFun == fun) {
             return;
         }
-        if (mCompiledDestructorFun && mCompiledDestructorFun != (destructor_fun_type)destroyClassInstance) {
+        if (mCompiledDestructorFun) {
             throw std::runtime_error("Can't change the compiled destructor!");
         }
 
@@ -273,8 +283,6 @@ public:
     int64_t mInitializationBitByteCount;
 };
 
-typedef VTable* vtable_ptr;
-
 //a class held directly inside of another object
 class HeldClass : public Type {
 public:
@@ -297,16 +305,60 @@ public:
             m_own_staticFunctions(staticFunctions),
             m_own_propertyFunctions(propertyFunctions),
             m_own_classMembers(classMembers),
-            m_hasComparisonOperators(false)
+            m_hasComparisonOperators(false),
+            m_hasGetAttributeMagicMethod(false),
+            m_refToType(nullptr)
     {
         m_name = inName;
+    }
 
-        if (m_memberFunctions.find("__eq__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
-        if (m_memberFunctions.find("__ne__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
-        if (m_memberFunctions.find("__lt__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
-        if (m_memberFunctions.find("__gt__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
-        if (m_memberFunctions.find("__le__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
-        if (m_memberFunctions.find("__ge__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+    ShaHash _computeIdentityHash(MutuallyRecursiveTypeGroup* groupHead = nullptr) {
+        ShaHash res = ShaHash(1, m_typeCategory) + ShaHash(m_name);
+
+        res += ShaHash(0);
+        for (auto b: m_bases) {
+            res += b->identityHash(groupHead);
+        }
+
+        res += ShaHash(1);
+        for (auto tup: m_own_members) {
+            res += ShaHash(std::get<0>(tup));
+            res += ShaHash(std::get<1>(tup)->identityHash(groupHead));
+            res += MutuallyRecursiveTypeGroup::tpInstanceShaHash(std::get<2>(tup), groupHead);
+        }
+
+        res += ShaHash(2);
+        for (auto nameAndFun: m_own_memberFunctions) {
+            res += ShaHash(nameAndFun.first);
+            res += nameAndFun.second->identityHash(groupHead);
+        }
+
+        res += ShaHash(3);
+        for (auto nameAndFun: m_own_staticFunctions) {
+            res += ShaHash(nameAndFun.first);
+            res += nameAndFun.second->identityHash(groupHead);
+        }
+
+        res += ShaHash(4);
+        for (auto nameAndFun: m_own_propertyFunctions) {
+            res += ShaHash(nameAndFun.first);
+            res += nameAndFun.second->identityHash(groupHead);
+        }
+
+        res += ShaHash(5);
+        for (auto nameAndFun: m_own_classMembers) {
+            res += ShaHash(nameAndFun.first);
+            res += MutuallyRecursiveTypeGroup::pyObjectShaHash(nameAndFun.second, groupHead);
+        }
+
+        return res;
+    }
+
+    template<class visitor_type>
+    void _visitCompilerVisibleInstances(const visitor_type& visitor) {
+        for (auto tup: m_own_members) {
+            visitor(std::get<2>(tup));
+        }
     }
 
     bool isBinaryCompatibleWithConcrete(Type* other);
@@ -320,6 +372,31 @@ public:
 
     template<class visitor_type>
     void _visitReferencedTypes(const visitor_type& visitor) {
+        for (auto& b: m_bases) {
+            Type* baseT = b;
+            visitor(baseT);
+            if (b != baseT) {
+                throw std::runtime_error("Somehow, we modified the base type of a HeldClass?");
+            }
+        }
+        for (auto& o: m_own_members) {
+            visitor(std::get<1>(o));
+        }
+        for (auto& o: m_own_memberFunctions) {
+            Type* t = std::get<1>(o);
+            visitor(t);
+            assert(t == std::get<1>(o));
+        }
+        for (auto& o: m_own_staticFunctions) {
+            Type* t = std::get<1>(o);
+            visitor(t);
+            assert(t == std::get<1>(o));
+        }
+        for (auto& o: m_own_propertyFunctions) {
+            Type* t = std::get<1>(o);
+            visitor(t);
+            assert(t == std::get<1>(o));
+        }
         for (auto& o: m_members) {
             visitor(std::get<1>(o));
         }
@@ -333,59 +410,32 @@ public:
             visitor(t);
             assert(t == std::get<1>(o));
         }
+        for (auto& o: m_propertyFunctions) {
+            Type* t = std::get<1>(o);
+            visitor(t);
+            assert(t == std::get<1>(o));
+        }
+    }
+
+    template<class visitor_type>
+    void _visitCompilerVisiblePythonObjects(const visitor_type& visitor) {
+        for (auto nameAndFun: m_own_classMembers) {
+            visitor(nameAndFun.second);
+        }
     }
 
     bool _updateAfterForwardTypesChanged();
 
     static HeldClass* Make(
-            std::string inName,
-            const std::vector<HeldClass*>& bases,
-            bool isFinal,
-            const std::vector<std::tuple<std::string, Type*, Instance> >& members,
-            const std::map<std::string, Function*>& memberFunctions,
-            const std::map<std::string, Function*>& staticFunctions,
-            const std::map<std::string, Function*>& propertyFunctions,
-            const std::map<std::string, PyObject*>& classMembers
-            )
-    {
-        //we only allow one base class to have members because we want native code to be
-        //able to just find those values in subclasses without hitting the vtable.
-        long countWithMembers = 0;
-
-        for (auto base: bases) {
-            if (base->m_members.size()) {
-                countWithMembers++;
-            }
-
-            if (base->isFinal()) {
-                throw std::runtime_error("Can't subclass " + base->name() + " because it's marked 'final'.");
-            }
-        }
-
-        if (countWithMembers > 1) {
-            throw std::runtime_error("Can't inherit from multiple base classes that both have members.");
-        }
-
-        HeldClass* result = new HeldClass(
-            inName,
-            bases,
-            isFinal,
-            members,
-            memberFunctions,
-            staticFunctions,
-            propertyFunctions,
-            classMembers
-        );
-
-        // we do these outside of the constructor so that if they throw we
-        // don't destroy the HeldClass type object (and just leak it instead) because
-        // we need to ensure we never delete Type objects.
-        result->initializeMRO();
-
-        result->endOfConstructorInitialization();
-
-        return result;
-    }
+        std::string inName,
+        const std::vector<HeldClass*>& bases,
+        bool isFinal,
+        const std::vector<std::tuple<std::string, Type*, Instance> >& members,
+        const std::map<std::string, Function*>& memberFunctions,
+        const std::map<std::string, Function*>& staticFunctions,
+        const std::map<std::string, Function*>& propertyFunctions,
+        const std::map<std::string, PyObject*>& classMembers
+    );
 
     // this gets called by Class. These types are always produced in pairs.
     void setClassType(Class* inClass) {
@@ -399,25 +449,23 @@ public:
         return m_classType;
     }
 
+    RefTo* getRefToType();
+
     HeldClass* renamed(std::string newName) {
         return Make(
             newName,
             m_bases,
             m_is_final,
-            m_members,
-            m_memberFunctions,
-            m_staticFunctions,
-            m_propertyFunctions,
-            m_classMembers
+            m_own_members,
+            m_own_memberFunctions,
+            m_own_staticFunctions,
+            m_own_propertyFunctions,
+            m_own_classMembers
         );
     }
 
-    // HeldClass is laid out as a VTable ptr, a set of member initialization fields, and then
+    // HeldClass is laid out as a set of member initialization fields, and then
     // the actual members.
-    static vtable_ptr& vtableFor(instance_ptr self) {
-        return *(vtable_ptr*)self;
-    }
-
     instance_ptr eltPtr(instance_ptr self, int64_t ix) const {
         return self + m_byte_offsets[ix];
     }
@@ -426,8 +474,6 @@ public:
 
     template<class buf_t>
     void deserialize(instance_ptr self, buf_t& buffer, size_t wireType) {
-        vtableFor(self) = m_vtable;
-
         for (long k = 0; k < m_members.size();k++) {
             clearInitializationFlag(self, k);
         }
@@ -456,14 +502,12 @@ public:
         buffer.writeEndCompound();
     }
 
-    void repr(instance_ptr self, ReprAccumulator& stream);
+    void repr(instance_ptr self, ReprAccumulator& stream, bool isStr, bool isClassNotHeldClass=false);
 
     typed_python_hash_type hash(instance_ptr left);
 
     template<class sub_constructor>
     void constructor(instance_ptr self, const sub_constructor& initializer) const {
-        vtableFor(self) = m_vtable;
-
         for (int64_t k = 0; k < m_members.size(); k++) {
             try {
                 initializer(eltPtr(self, k), k);
@@ -479,8 +523,6 @@ public:
 
     void setAttribute(instance_ptr self, int memberIndex, instance_ptr other) const;
 
-    void emptyConstructor(instance_ptr self);
-
     //don't default construct classes
     static bool wantsToDefaultConstruct(Type* t) {
         return t->is_default_constructible() && t->getTypeCategory() != TypeCategory::catClass;
@@ -495,7 +537,7 @@ public:
     void assign(instance_ptr self, instance_ptr other);
 
     bool checkInitializationFlag(instance_ptr self, int memberIndex) const {
-        int byte = memberIndex / 8 + sizeof(vtable_ptr);
+        int byte = memberIndex / 8;
         int bit = memberIndex % 8;
         return bool( ((uint8_t*)self)[byte] & (1 << bit) );
     }
@@ -532,6 +574,10 @@ public:
         return m_members;
     }
 
+    const std::vector<std::tuple<std::string, Type*, Instance> >& getOwnMembers() const {
+        return m_own_members;
+    }
+
     int getMemberIndex(const char* attrName) const {
         auto it = m_membersByName.find(attrName);
 
@@ -558,6 +604,22 @@ public:
         return m_propertyFunctions;
     }
 
+    const std::map<std::string, Function*>& getOwnMemberFunctions() const {
+        return m_own_memberFunctions;
+    }
+
+    const std::map<std::string, Function*>& getOwnStaticFunctions() const {
+        return m_own_staticFunctions;
+    }
+
+    const std::map<std::string, PyObject*>& getOwnClassMembers() const {
+        return m_own_classMembers;
+    }
+
+    const std::map<std::string, Function*>& getOwnPropertyFunctions() const {
+        return m_own_propertyFunctions;
+    }
+
     const std::vector<size_t>& getOffsets() const {
         return m_byte_offsets;
     }
@@ -568,6 +630,10 @@ public:
         return m_hasComparisonOperators;
     }
 
+    bool hasGetAttributeMagicMethod() const {
+        return m_hasGetAttributeMagicMethod;
+    }
+
     void initializeMRO() {
         // this is not how the MRO actually works, but we have yet to actually
         // code it correctly.
@@ -580,8 +646,6 @@ public:
                 m_mro.push_back(cls);
                 m_ancestor_to_mro_index[cls] = m_mro.size() - 1;
 
-                m_bases_as_set.insert(cls);
-
                 for (auto b: cls->getBases()) {
                     visit(b);
                 }
@@ -646,16 +710,29 @@ public:
             ancestor->m_implementors.insert(this);
         }
 
-        m_vtable->finalize(&mClassDispatchTables[0], (m_members.size() + 7) / 8);
+        m_vtable->finalize(&mClassDispatchTables[0], (m_members.size() + 7) / 8, mClassDispatchTables.size());
 
         // make sure that, for every interface we can take on, we have slots allocated
         // that the compiler can come along and compile.
         for (HeldClass* ancestor: m_mro) {
             dispatchTableAs(ancestor)->initialize(ancestor->dispatchTableAs(ancestor));
         }
+
+        setMagicMethodExistConstants();
     }
 
-    size_t allocateMethodDispatch(std::string funcName, function_signature_type signature) {
+    void setMagicMethodExistConstants() {
+        if (m_memberFunctions.find("__eq__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+        if (m_memberFunctions.find("__ne__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+        if (m_memberFunctions.find("__lt__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+        if (m_memberFunctions.find("__gt__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+        if (m_memberFunctions.find("__le__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+        if (m_memberFunctions.find("__ge__") != m_memberFunctions.end()) { m_hasComparisonOperators = true; }
+
+        if (m_memberFunctions.find("__getattribute__") != m_memberFunctions.end()) { m_hasGetAttributeMagicMethod = true; }
+    }
+
+    size_t allocateMethodDispatch(std::string funcName, function_call_signature_type signature) {
         size_t result = dispatchTableAs(this)->allocateMethodDispatch(funcName, signature);
 
         // make sure we add this dispatch to every child that implements us as an interface
@@ -709,9 +786,14 @@ public:
         }
     }
 
-
     bool isSubclassOfConcrete(Type* otherType) {
-        return m_bases_as_set.find((HeldClass*)otherType) != m_bases_as_set.end();
+        for (auto t: m_mro) {
+            if (Type::typesEquivalent(t, otherType)) {
+                return true;
+            }
+        }
+
+        return false;
     }
 
     VTable* getVTable() const {
@@ -725,11 +807,11 @@ public:
     int64_t getMroIndex(HeldClass* ancestor) const {
         auto it = m_ancestor_to_mro_index.find(ancestor);
 
-        if (it == m_ancestor_to_mro_index.end()) {
-            return -1;
+        if (it != m_ancestor_to_mro_index.end()) {
+            return it->second;
         }
 
-        return it->second;
+        return -1;
     }
 
     ClassDispatchTable* dispatchTableAs(HeldClass* interface) {
@@ -741,7 +823,10 @@ public:
         return &mClassDispatchTables[offset];
     }
 
-    BoundMethod* getMemberFunctionMethodType(const char* attr);
+    // get a BoundMethod type object representing this method on instances
+    // of this class. If 'forHeld', then 'self' will be bound as a RefTo(Held(T))
+    // instead of a T
+    BoundMethod* getMemberFunctionMethodType(const char* attr, bool forHeld);
 
 private:
     std::vector<size_t> m_byte_offsets;
@@ -759,12 +844,12 @@ private:
 
     std::unordered_map<HeldClass*, int> m_ancestor_to_mro_index;
 
-    std::unordered_set<HeldClass*> m_bases_as_set;
-
     VTable* m_vtable;
 
     Class* m_classType; //the non-held version of this class
 
+    RefTo* m_refToType; //a cache for the ref-to-type
+
     std::vector<ClassDispatchTable> mClassDispatchTables;
 
     //the members we
@@ -791,9 +876,8 @@ private:
 
     std::unordered_map<const char*, size_t, HashConstCharPtr, ConstCharPtrsAreEqual> m_membersByName;
 
-    std::unordered_map<const char*, BoundMethod*, HashConstCharPtr, ConstCharPtrsAreEqual> m_memberFunctionMethodTypes;
+    std::unordered_map<const char*, BoundMethod*, HashConstCharPtr, ConstCharPtrsAreEqual> m_memberFunctionMethodTypes[2];
 
     bool m_hasComparisonOperators;
-
+    bool m_hasGetAttributeMagicMethod;
 };
-
