@@ -22,7 +22,12 @@ class PyHeldClassInstance : public PyInstance {
 public:
     typedef HeldClass modeled_type;
 
-    static bool pyValCouldBeOfTypeConcrete(modeled_type* type, PyObject* pyRepresentation, bool isExplicit) {
+    static bool pyValCouldBeOfTypeConcrete(modeled_type* type, PyObject* pyRepresentation, ConversionLevel level) {
         return true;
     }
+
+    static void mirrorTypeInformationIntoPyTypeConcrete(HeldClass* classT, PyTypeObject* pyType) {
+        //expose 'ElementType' as a member of the type object
+        PyDict_SetItemString(pyType->tp_dict, "Class", typePtrToPyTypeRepresentation(classT->getClassType()));
+    }
 };
