@@ -18,6 +18,7 @@
 
 #include "util.hpp"
 #include "AllTypes.hpp"
+#include "ConversionLevel.hpp"
 
 //extension of PyTypeObject that adds a Type* at the end.
 struct NativeTypeWrapper {
@@ -38,9 +39,11 @@ class PyTupleOfInstance;
 class PyDictInstance;
 class PyConstDictInstance;
 class PyPointerToInstance;
+class PyRefToInstance;
 class PyCompositeTypeInstance;
 class PyTupleInstance;
 class PyNamedTupleInstance;
+class PyAlternativeMatcherInstance;
 class PyAlternativeInstance;
 class PyConcreteAlternativeInstance;
 class PyStringInstance;
@@ -55,6 +58,8 @@ class PyOneOfInstance;
 class PyForwardInstance;
 class PyEmbeddedMessageInstance;
 class PySetInstance;
+class PyPyCellInstance;
+class PyTypedCellInstance;
 
 template<class T>
 class PyRegisterTypeInstance;
@@ -64,9 +69,9 @@ public:
     PyObject_HEAD
 
     bool mIsInitialized;
-    bool mIsMatcher;
     char mIteratorFlag; //0 is keys, 1 is values, 2 is pairs
     int64_t mIteratorOffset; //-1 if we're not an iterator
+    int64_t mContainerSize; //-1 if we're not an iterator
 
     Instance mContainingInstance;
 
@@ -101,6 +106,8 @@ public:
                 return f(*(PyTupleOfInstance*)obj);
             case Type::TypeCategory::catPointerTo:
                  return f(*(PyPointerToInstance*)obj);
+            case Type::TypeCategory::catRefTo:
+                 return f(*(PyRefToInstance*)obj);
             case Type::TypeCategory::catListOf:
                 return f(*(PyListOfInstance*)obj);
             case Type::TypeCategory::catNamedTuple:
@@ -117,6 +124,8 @@ public:
                 return f(*(PyAlternativeInstance*)obj);
             case Type::TypeCategory::catConcreteAlternative:
                 return f(*(PyConcreteAlternativeInstance*)obj);
+            case Type::TypeCategory::catAlternativeMatcher:
+                return f(*(PyAlternativeMatcherInstance*)obj);
             case Type::TypeCategory::catPythonSubclass:
                 return f(*(PyPythonSubclassInstance*)obj);
             case Type::TypeCategory::catPythonObjectOfType:
@@ -131,6 +140,10 @@ public:
                 return f(*(PyBoundMethodInstance*)obj);
             case Type::TypeCategory::catEmbeddedMessage:
                 return f(*(PyEmbeddedMessageInstance*)obj);
+            case Type::TypeCategory::catPyCell:
+                return f(*(PyPyCellInstance*)obj);
+            case Type::TypeCategory::catTypedCell:
+                return f(*(PyTypedCellInstance*)obj);
             case Type::TypeCategory::catNone:
             case Type::TypeCategory::catInt64:
             case Type::TypeCategory::catFloat64:
@@ -176,6 +189,8 @@ public:
                 return f((PyTupleOfInstance*)nullptr);
             case Type::TypeCategory::catPointerTo:
                  return f((PyPointerToInstance*)nullptr);
+            case Type::TypeCategory::catRefTo:
+                 return f((PyRefToInstance*)nullptr);
             case Type::TypeCategory::catListOf:
                 return f((PyListOfInstance*)nullptr);
             case Type::TypeCategory::catNamedTuple:
@@ -202,6 +217,8 @@ public:
                 return f((PyHeldClassInstance*)nullptr);
             case Type::TypeCategory::catFunction:
                 return f((PyFunctionInstance*)nullptr);
+            case Type::TypeCategory::catAlternativeMatcher:
+                return f((PyAlternativeMatcherInstance*)nullptr);
             case Type::TypeCategory::catBoundMethod:
                 return f((PyBoundMethodInstance*)nullptr);
             case Type::TypeCategory::catNone:
@@ -212,6 +229,10 @@ public:
                 return f((PyBytesInstance*)nullptr);
             case Type::TypeCategory::catEmbeddedMessage:
                 return f((PyEmbeddedMessageInstance*)nullptr);
+            case Type::TypeCategory::catPyCell:
+                return f((PyPyCellInstance*)nullptr);
+            case Type::TypeCategory::catTypedCell:
+                return f((PyTypedCellInstance*)nullptr);
             case Type::TypeCategory::catOneOf:
                 return f((PyOneOfInstance*)nullptr);
             case Type::TypeCategory::catForward:
@@ -224,9 +245,9 @@ public:
     static int reversePyOpOrdering(int op);
 
     template<class T>
-    static int specializeForTypeReturningInt(PyObject* obj, const T& f) {
+    static int specializeForTypeReturningInt(PyObject* obj, const T& f, Type* typeOverride = nullptr) {
         try {
-            return specialize(obj, f);
+            return specialize(obj, f, typeOverride);
         } catch(PythonExceptionSet& e) {
             return -1;
         } catch(std::exception& e) {
@@ -272,11 +293,13 @@ public:
     //rather than the standard python representation.
     template<class init_func>
     static PyObject* initialize(Type* eltType, const init_func& f) {
-        eltType->assertForwardsResolved();
+        eltType->assertForwardsResolvedSufficientlyToInstantiate();
 
         PyInstance* self =
             (PyInstance*)typeObj(eltType)->tp_alloc(typeObj(eltType), 0);
 
+        self->mIteratorOffset = -1;
+
         try {
             self->initialize(f, eltType);
 
@@ -290,7 +313,7 @@ public:
     template<class init_func>
     void initialize(const init_func& i, Type* typeIfKnown = nullptr) {
         Type* type = typeIfKnown ? typeIfKnown : extractTypeFrom(((PyObject*)this)->ob_type);
-        type->assertForwardsResolved();
+        type->assertForwardsResolvedSufficientlyToInstantiate();
 
         mIsInitialized = false;
         new (&mContainingInstance) Instance( type, i );
@@ -307,14 +330,14 @@ public:
         });
     }
 
-    PyObject* createIteratorToSelf(int32_t iterTypeFlag) {
+    PyObject* createIteratorToSelf(int32_t iterTypeFlag, int64_t containerSize) {
         PyInstance* result = (PyInstance*)initialize(type(), [&](instance_ptr out) {
             type()->copy_constructor(out, dataPtr());
         });
 
-        result->mIsMatcher = false;
         result->mIteratorOffset = 0;
         result->mIteratorFlag = iterTypeFlag;
+        result->mContainerSize = containerSize;
 
         return (PyObject*)result;
     }
@@ -323,22 +346,24 @@ public:
 
     Type* type();
 
+    std::pair<Type*, instance_ptr> derefAnyRefTo();
+
     static PyMethodDef* typeMethods(Type* t);
 
     static PyMethodDef* typeMethodsConcrete(Type* t);
 
     static void tp_dealloc(PyObject* self);
 
-    static bool pyValCouldBeOfType(Type* t, PyObject* pyRepresentation, bool isExplicit);
+    static bool pyValCouldBeOfType(Type* t, PyObject* pyRepresentation, ConversionLevel level);
 
     /**
      construct an 'eltType' from a python object at 'tgt'. If 'isExplicit' then we're invoked from an explicit
      copy constructor, so more liberal conversion is allowed than if 'isExplicit' is false, which happens
      when we're attempting to convert for purposes of method dispatch.
      */
-    static void copyConstructFromPythonInstance(Type* eltType, instance_ptr tgt, PyObject* pyRepresentation, bool isExplicit=false);
+    static void copyConstructFromPythonInstance(Type* eltType, instance_ptr tgt, PyObject* pyRepresentation, ConversionLevel level);
 
-    static void copyConstructFromPythonInstanceConcrete(Type* eltType, instance_ptr tgt, PyObject* pyRepresentation, bool isExplicit);
+    static void copyConstructFromPythonInstanceConcrete(Type* eltType, instance_ptr tgt, PyObject* pyRepresentation, ConversionLevel level);
 
     static void constructFromPythonArguments(uint8_t* data, Type* t, PyObject* args, PyObject* kwargs);
 
@@ -462,6 +487,8 @@ public:
 
     static PyObject* undefinedBehaviorException();
 
+    static PyObject* nonTypesAcceptedAsTypes();
+
     static PySequenceMethods* sequenceMethodsFor(Type* t);
 
     static PyNumberMethods* numberMethods(Type* t);
@@ -492,6 +519,11 @@ public:
 
     static Type* extractTypeFrom(PyTypeObject* typeObj);
 
+    // if 'obj' is a PyInstance, return its type and data ptrs, after unwrapping
+    // any 'RefTo' classes, which is the standard behavior for most classes,
+    // which don't specifically know anything about 'RefTo'
+    static std::pair<Type*, instance_ptr> extractTypeAndPtrFrom(PyObject* obj);
+
     static int tp_setattro(PyObject *o, PyObject* attrName, PyObject* attrVal);
 
     int tp_setattr_concrete(PyObject* attrName, PyObject* attrVal);
@@ -514,6 +546,8 @@ public:
 
     static bool compare_to_python_concrete(Type* t, instance_ptr self, PyObject* other, bool exact, int pyComparisonOp);
 
+    bool compare_as_iterator_to_python_concrete(PyObject* other, int pyComparisonOp);
+
     static PyObject *tp_richcompare(PyObject *a, PyObject *b, int op);
 
     static PyObject* tp_iter(PyObject *o);
@@ -526,8 +560,14 @@ public:
 
     static PyObject* tp_repr(PyObject *o);
 
+    PyObject* tp_repr_concrete();
+
+    static PyObject* tp_descr_get(PyObject* func, PyObject* obj, PyObject* type);
+
     static PyObject* tp_str(PyObject *o);
 
+    PyObject* tp_str_concrete();
+
     static bool typeCanBeSubclassed(Type* t);
 
     static PyBufferProcs* bufferProcs();
