@@ -1,5 +1,5 @@
 /******************************************************************************
-   Copyright 2017-2019 typed_python Authors
+   Copyright 2017-2020 typed_python Authors
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
@@ -28,17 +28,41 @@ ListOfType* PyListOfInstance::type() {
     return (ListOfType*)extractTypeFrom(((PyObject*)this)->ob_type);
 }
 
-bool PyTupleOrListOfInstance::pyValCouldBeOfTypeConcrete(modeled_type* type, PyObject* pyRepresentation, bool isExplicit) {
-    if (!isExplicit) {
-        return PyList_Check(pyRepresentation) || PyTuple_Check(pyRepresentation);
+bool PyTupleOrListOfInstance::pyValCouldBeOfTypeConcrete(modeled_type* type, PyObject* pyRepresentation, ConversionLevel level) {
+    if (type->isListOf()) {
+        return level >= ConversionLevel::ImplicitContainers;
     }
 
-    return
-        PyTuple_Check(pyRepresentation) ||
-        PyList_Check(pyRepresentation) ||
-        PyDict_Check(pyRepresentation) ||
-        PyIter_Check(pyRepresentation)
-        ;
+    std::pair<Type*, instance_ptr> typeAndPtrOfArg = extractTypeAndPtrFrom(pyRepresentation);
+
+    if (level == ConversionLevel::Signature) {
+        return Type::typesEquivalent(typeAndPtrOfArg.first, type);
+    }
+
+    if (level == ConversionLevel::Upcast) {
+        return false;
+    }
+
+    if (level >= ConversionLevel::ImplicitContainers) {
+        return true;
+    }
+
+    // only allow implicit conversion from tuple/list of. We don't want dicts and sets to implicitly
+    // convert to tuples
+    if (PyTuple_Check(pyRepresentation) || PyList_Check(pyRepresentation)) {
+        return true;
+    }
+
+    if (typeAndPtrOfArg.first && (
+            typeAndPtrOfArg.first->isListOf()
+            || typeAndPtrOfArg.first->isTupleOf()
+            || typeAndPtrOfArg.first->isComposite()
+            )
+    ) {
+        return true;
+    }
+
+    return false;
 }
 
 PyObject* PyTupleOrListOfInstance::pyOperatorConcreteReverse(PyObject* lhs, const char* op, const char* opErrRep) {
@@ -93,7 +117,11 @@ PyObject* PyTupleOrListOfInstance::pyOperatorAdd(PyObject* rhs, const char* op,
     }
 
     //generic path to add any kind of iterable.
-    if (PyObject_Length(rhs) != -1) {
+    if (PyObject_Length(rhs) == -1) {
+        // if not iterable, we want to fall through to Py_NotImplemented, which will try reverse operator,
+        // instead of generating an actual error
+        PyErr_Clear();
+    } else {
         Type* eltType = type()->getEltType();
 
         return PyInstance::initialize(type(), [&](instance_ptr data) {
@@ -115,23 +143,24 @@ PyObject* PyTupleOrListOfInstance::pyOperatorAdd(PyObject* rhs, const char* op,
                             throw PythonExceptionSet();
                         }
 
-                        PyInstance::copyConstructFromPythonInstance(eltType, eltPtr, o, true);
+                        PyInstance::copyConstructFromPythonInstance(eltType, eltPtr, o, ConversionLevel::ImplicitContainers);
                     }
                 });
         });
     }
 
-    PyErr_SetString(
-        PyExc_TypeError,
-        (std::string("cannot concatenate ") + type()->name() + " and "
-                + rhs->ob_type->tp_name).c_str()
-        );
-
-    return NULL;
+    return PyInstance::pyOperatorConcrete(rhs, op, opErr);
 }
 
 
 //static
+PyDoc_STRVAR(listSetSizeUnsafe_doc,
+    "lst.setSizeUnsafe(n) -> None, and set length of lst to n\n"
+    "\n"
+    "No adjustments to elements or reallocations are done.\n"
+    "The size field is simply set to n."
+    "This is intended for compiled code.\n"
+    );
 PyObject* PyListOfInstance::listSetSizeUnsafe(PyObject* o, PyObject* args) {
     PyListOfInstance* self_w = (PyListOfInstance*)o;
 
@@ -192,14 +221,8 @@ bool constructTupleOrListInstFromNumpy(TupleOrListOfType* tupT, instance_ptr tgt
 
     return true;
 }
-void PyTupleOrListOfInstance::copyConstructFromPythonInstanceConcrete(TupleOrListOfType* tupT, instance_ptr tgt, PyObject* pyRepresentation, bool isExplicit) {
-    if (!isExplicit) {
-        throw std::logic_error("Can't convert from " +
-            std::string(pyRepresentation->ob_type->tp_name) + " to " + tupT->name()
-        );
-    }
-
-    if (PyArray_Check(pyRepresentation)) {
+void PyTupleOrListOfInstance::copyConstructFromPythonInstanceConcrete(TupleOrListOfType* tupT, instance_ptr tgt, PyObject* pyRepresentation, ConversionLevel level) {
+    if (PyArray_Check(pyRepresentation) && level >= ConversionLevel::ImplicitContainers) {
         if (!PyArray_ISBEHAVED_RO(pyRepresentation)) {
             throw std::logic_error("Can't convert a numpy array that's not contiguous and in machine-native byte order.");
         }
@@ -268,43 +291,61 @@ void PyTupleOrListOfInstance::copyConstructFromPythonInstanceConcrete(TupleOrLis
         }
     }
 
-    if (PyTuple_Check(pyRepresentation)) {
-        tupT->constructor(tgt, PyTuple_Size(pyRepresentation),
-            [&](uint8_t* eltPtr, int64_t k) {
-                PyObjectHolder arg(PyTuple_GetItem(pyRepresentation,k));
-                PyInstance::copyConstructFromPythonInstance(tupT->getEltType(), eltPtr, arg, isExplicit);
-                }
-            );
-        return;
-    }
-    if (PyList_Check(pyRepresentation)) {
-        tupT->constructor(tgt, PyList_Size(pyRepresentation),
-            [&](uint8_t* eltPtr, int64_t k) {
-                PyObjectHolder listItem(PyList_GetItem(pyRepresentation,k));
-                PyInstance::copyConstructFromPythonInstance(tupT->getEltType(), eltPtr, listItem, isExplicit);
-                }
-            );
-        return;
-    }
+    ConversionLevel childLevel = ConversionLevel::ImplicitContainers;
 
-    if (PySet_Check(pyRepresentation)) {
-        if (PySet_Size(pyRepresentation) == 0) {
-            tupT->constructor(tgt);
-            return;
+    // determine the level of conversion we'll use when converting the children if we're not a New call
+    if (level < ConversionLevel::ImplicitContainers) {
+        // ListOf only operates if we're above ImplicitContainers
+        if (tupT->isListOf()) {
+            throw std::logic_error("Can't implicitly convert from " +
+                std::string(pyRepresentation->ob_type->tp_name) + " to " + tupT->name()
+            );
         }
 
-        PyObjectStealer iterator(PyObject_GetIter(pyRepresentation));
+        if (level == ConversionLevel::Upcast) {
+            throw std::logic_error("Can't upcast from " +
+                std::string(pyRepresentation->ob_type->tp_name) + " to " + tupT->name()
+            );
+        } else if (level == ConversionLevel::UpcastContainers) {
+            childLevel = level;
+        } else if (level == ConversionLevel::Signature) {
+            PyInstance::copyConstructFromPythonInstanceConcrete(tupT, tgt, pyRepresentation, level);
+            return;
+        }
+    }
 
-        tupT->constructor(tgt, PySet_Size(pyRepresentation),
-            [&](uint8_t* eltPtr, int64_t k) {
-                PyObjectStealer item(PyIter_Next(iterator));
+    if (level < ConversionLevel::ImplicitContainers) {
+        // only allow implicit conversion from tuple/list of. We don't want dicts and sets to implicitly
+        // convert to tuples
+        bool isValid = false;
+        if (PyTuple_Check(pyRepresentation) || PyList_Check(pyRepresentation)) {
+            isValid = true;
+        }
 
-                if (!item) {
-                    throw std::logic_error("Set ran out of elements.");
-                }
+        std::pair<Type*, instance_ptr> typeAndPtrOfArg = extractTypeAndPtrFrom(pyRepresentation);
+        if (typeAndPtrOfArg.first && (
+                typeAndPtrOfArg.first->isListOf()
+                || typeAndPtrOfArg.first->isTupleOf()
+                || typeAndPtrOfArg.first->isComposite())
+        ) {
+            isValid = true;
+        }
 
-                PyInstance::copyConstructFromPythonInstance(tupT->getEltType(), eltPtr, item, isExplicit);
-            });
+        if (!isValid) {
+            PyInstance::copyConstructFromPythonInstanceConcrete(tupT, tgt, pyRepresentation, level);
+            return;
+        }
+    }
+
+    if (PyBytes_Check(pyRepresentation) && tupT->getEltType()->getTypeCategory() == Type::TypeCategory::catUInt8) {
+        tupT->constructor(tgt);
+
+        size_t sz = PyBytes_GET_SIZE(pyRepresentation);
+
+        tupT->reserve(tgt, sz);
+
+        memcpy(tupT->eltPtr(tgt, 0), PyBytes_AsString(pyRepresentation), sz);
+        tupT->setSizeUnsafe(tgt, sz);
 
         return;
     }
@@ -324,7 +365,7 @@ void PyTupleOrListOfInstance::copyConstructFromPythonInstanceConcrete(TupleOrLis
                     return false;
                 }
 
-                PyInstance::copyConstructFromPythonInstance(tupT->getEltType(), eltPtr, item, isExplicit);
+                PyInstance::copyConstructFromPythonInstance(tupT->getEltType(), eltPtr, item, childLevel);
 
                 return true;
             });
@@ -335,6 +376,93 @@ void PyTupleOrListOfInstance::copyConstructFromPythonInstanceConcrete(TupleOrLis
     }
 }
 
+PyObject* PyTupleOrListOfInstance::toBytes(PyObject* o, PyObject* args) {
+    PyListOfInstance* self_w = (PyListOfInstance*)o;
+
+    if (!self_w->type()->getEltType()->isPOD()) {
+        PyErr_Format(
+            PyExc_TypeError,
+            "Can't convert %s to bytes because internals are not POD",
+            self_w->type()->name().c_str()
+        );
+
+        return NULL;
+    }
+
+    return PyBytes_FromStringAndSize(
+        (const char*)self_w->type()->eltPtr(self_w->dataPtr(), 0),
+        self_w->type()->count(self_w->dataPtr()) *
+        self_w->type()->getEltType()->bytecount()
+    );
+}
+
+PyObject* PyTupleOrListOfInstance::fromBytes(PyObject* o, PyObject* args, PyObject* kwds) {
+    static const char *kwlist[] = {"bytesObj", NULL};
+
+    PyObject* bytesObj;
+
+    if (!PyArg_ParseTupleAndKeywords(args, NULL, "O", (char**)kwlist, &bytesObj)) {
+        return nullptr;
+    }
+
+    if (!PyBytes_Check(bytesObj)) {
+        PyErr_Format(
+            PyExc_TypeError,
+            "Expected 'bytesObj' to be a bytes object. Got %s",
+            bytesObj->ob_type->tp_name
+        );
+    }
+
+    Type* selfType = PyInstance::unwrapTypeArgToTypePtr(o);
+
+    if (!selfType || !selfType->isTupleOrListOf()) {
+        PyErr_Format(PyExc_TypeError, "Expected cls to be a Type");
+        return nullptr;
+    }
+
+    TupleOrListOfType* tupT = (TupleOrListOfType*)selfType;
+
+    if (!tupT->getEltType()->isPOD()) {
+        PyErr_Format(
+            PyExc_TypeError,
+            "Can't convert %s from bytes because internals are not POD",
+            tupT->name().c_str()
+        );
+        return nullptr;
+    }
+
+    int64_t bytecount = tupT->getEltType()->bytecount();
+    size_t sz = PyBytes_GET_SIZE(bytesObj);
+
+    if ((bytecount == 0 && sz) || sz % bytecount != 0) {
+        PyErr_Format(PyExc_ValueError, "Byte array must be an integer multiple of underlying type.");
+        return nullptr;
+    }
+
+    size_t eltCount = bytecount > 0 ? sz / bytecount : 0;
+
+    Instance outConverted(selfType, [&](instance_ptr data) {
+        tupT->constructor(data);
+        tupT->reserve(data, eltCount);
+        memcpy(tupT->eltPtr(data, 0), PyBytes_AsString(bytesObj), sz);
+        tupT->setSizeUnsafe(data, eltCount);
+    });
+
+    return PyInstance::fromInstance(outConverted);
+}
+
+PyDoc_STRVAR(TupleOf_toArray_doc,
+    "t.toarray() -> numpy array\n"
+    "\n"
+    "Converts a TupleOf() instance to a numpy array\n"
+    "Raises TypeError on failure.\n"
+    );
+PyDoc_STRVAR(ListOf_toArray_doc,
+    "lst.toarray() -> numpy array\n"
+    "\n"
+    "Converts a ListOf() instance to a numpy array\n"
+    "Raises TypeError on failure.\n"
+    );
 PyObject* PyTupleOrListOfInstance::toArray(PyObject* o, PyObject* args) {
     PyListOfInstance* self_w = (PyListOfInstance*)o;
     npy_intp dims[1] = { self_w->type()->count(self_w->dataPtr()) };
@@ -482,20 +610,47 @@ PyObject* PyTupleOrListOfInstance::mp_subscript_concrete(PyObject* item) {
         return sq_item((PyObject*)this, PyLong_AsLongLong(res));
     }
 
-    PyErr_SetObject(PyExc_KeyError, item);
+    PyErr_Format(PyExc_TypeError, "ListOf indices must be integers or slices, not %S", item);
     return NULL;
 }
 
-PyMethodDef* PyTupleOfInstance::typeMethodsConcrete(Type* t) {
-    return new PyMethodDef [3] {
-        {"toArray", (PyCFunction)PyTupleOrListOfInstance::toArray, METH_VARARGS, NULL},
-        {NULL, NULL}
-    };
-}
+const char* TUPLE_CONVERT_DOCSTRING =
+    "Convert an arbitrary value to a `TupleOf(T)`\n\n"
+    "Unlike calling `TupleOf(T)`, `TupleOf(T).convert` will recursively attempt\n"
+    "to convert all of the interior values in the argument as well.\n"
+    "Normally, `TupleOf(T)([x])` will only succeed if `x` is implicitly convertible to `T`\n"
+    "TupleOf(T).convert([x]) will call 'convert' on each of the members of the iterable.\n"
+    "Note that this means that you may get deepcopies of objects: `TupleOf(TupleOf(int))([[x]])`\n"
+    "will duplicate the inner list as a `TupleOf(int)`."
+;
 
-//static
-PyObject* PyListOfInstance::listPointerUnsafe(PyObject* o, PyObject* args) {
-    PyListOfInstance* self_w = (PyListOfInstance*)o;
+
+const char* TUPLE_TO_BYTES_DOCSTRING =
+    "Convert a TupleOf(T) to the raw bytes underneath it.\n\n"
+    "This is only valid if T is \"POD\" (Plain Old Data), meaning it must be\n"
+    "an integer, float, or bool type, or some combination of those in a Tuple or \n"
+    "NamedTuple."
+;
+
+const char* TUPLE_FROM_BYTES_DOCSTRING =
+    "Construct a TupleOf(T) from the raw bytes that would underly it.\n\n"
+    "This is only valid if T is \"POD\" (Plain Old Data), meaning it must be\n"
+    "an integer, float, or bool type, or some combination of those in a Tuple or \n"
+    "NamedTuple."
+;
+
+PyDoc_STRVAR(tuplePointerUnsafe_doc,
+    "tup.pointerUnsafe(i) -> pointer to element i of tup\n"
+    "\n"
+    "If tup if of type TupleOf(T), the return value is of type PointerTo(T).\n"
+    "This is intended for compiled code, where pointerUnsafe is expected to be\n"
+    "faster than the usual bounds-checked indexed access to an instance of\n"
+    "TupleOf(T).\n"
+    "In C++ terms, tup.pointerUnsafe(i) is &tup[i].\n"
+
+);
+PyObject* PyTupleOrListOfInstance::pointerUnsafe(PyObject* o, PyObject* args) {
+    PyTupleOrListOfInstance* self_w = (PyTupleOrListOfInstance*)o;
 
     if (PyTuple_Size(args) != 1 || !PyLong_Check(PyTuple_GetItem(args, 0))) {
         PyErr_SetString(PyExc_TypeError, "ListOf.pointerUnsafe takes one integer argument");
@@ -509,16 +664,44 @@ PyObject* PyListOfInstance::listPointerUnsafe(PyObject* o, PyObject* args) {
     return extractPythonObject((instance_ptr)&ptr, PointerTo::Make(self_w->type()->getEltType()));
 }
 
+
+
+PyMethodDef* PyTupleOfInstance::typeMethodsConcrete(Type* t) {
+    return new PyMethodDef [6] {
+        {"toArray", (PyCFunction)PyTupleOrListOfInstance::toArray, METH_VARARGS, TupleOf_toArray_doc},
+        {"toBytes", (PyCFunction)PyTupleOrListOfInstance::toBytes, METH_VARARGS, TUPLE_TO_BYTES_DOCSTRING},
+        {"fromBytes", (PyCFunction)PyTupleOrListOfInstance::fromBytes, METH_VARARGS | METH_KEYWORDS | METH_CLASS, TUPLE_FROM_BYTES_DOCSTRING},
+        {"pointerUnsafe", (PyCFunction)PyTupleOrListOfInstance::pointerUnsafe, METH_VARARGS, tuplePointerUnsafe_doc},
+        {NULL, NULL}
+    };
+}
+
+//static
+PyDoc_STRVAR(listPointerUnsafe_doc,
+    "lst.pointerUnsafe(i) -> pointer to element i of lst\n"
+    "\n"
+    "If lst if of type ListOf(T), the return value is of type PointerTo(T).\n"
+    "This is intended for compiled code, where pointerUnsafe is expected to be\n"
+    "faster than the usual bounds-checked indexed access to an instance of\n"
+    "ListOf(T).\n"
+    "In C++ terms, lst.pointerUnsafe(i) is &lst[i].\n"
+);
+
 // static
+PyDoc_STRVAR(listAppend_doc,
+    "lst.append(e) -> None, and appends element e to lst"
+    );
 PyObject* PyListOfInstance::listAppend(PyObject* o, PyObject* args) {
+    if (PyTuple_Size(args) != 1) {
+        PyErr_SetString(PyExc_TypeError, "ListOf.append takes one argument");
+        return NULL;
+    }
+
+    return listAppendDirect(o, PyTuple_GetItem(args, 0));
+}
+
+PyObject* PyListOfInstance::listAppendDirect(PyObject* o, PyObject* value) {
     try {
-        if (PyTuple_Size(args) != 1) {
-            PyErr_SetString(PyExc_TypeError, "ListOf.append takes one argument");
-            return NULL;
-        }
-
-        PyObjectHolder value(PyTuple_GetItem(args, 0));
-
         PyListOfInstance* self_w = (PyListOfInstance*)o;
 
         Type* value_type = extractTypeFrom(value->ob_type);
@@ -531,7 +714,12 @@ PyObject* PyListOfInstance::listAppend(PyObject* o, PyObject* args) {
             self_w->type()->append(self_w->dataPtr(), value_w->dataPtr());
         } else {
             Instance temp(eltType, [&](instance_ptr data) {
-                PyInstance::copyConstructFromPythonInstance(eltType, data, value, true);
+                PyInstance::copyConstructFromPythonInstance(
+                    eltType,
+                    data,
+                    value,
+                    ConversionLevel::ImplicitContainers
+                );
             });
 
             self_w->type()->append(self_w->dataPtr(), temp.data());
@@ -546,7 +734,11 @@ PyObject* PyListOfInstance::listAppend(PyObject* o, PyObject* args) {
     }
 }
 
+
 // static
+PyDoc_STRVAR(listExtend_doc,
+    "lst.extend(c) -> None, and appends elements of the container c to lst"
+    );
 PyObject* PyListOfInstance::listExtend(PyObject* o, PyObject* args) {
     return translateExceptionToPyObject([&]() {
         if (PyTuple_Size(args) != 1) {
@@ -581,18 +773,24 @@ PyObject* PyListOfInstance::listExtend(PyObject* o, PyObject* args) {
             return incref(Py_None);
         }
 
-        // try to convert this to a list
-        Instance temp(self_type, [&](instance_ptr data) {
-            PyInstance::copyConstructFromPythonInstance(self_type, data, value, true);
+        // iterate
+        iterate(value, [&](PyObject* arg) {
+            if (!listAppendDirect(o, arg)) {
+                throw PythonExceptionSet();
+            }
         });
 
-        extendFromBinaryCompatiblePtr(temp.data());
-
         return incref(Py_None);
     });
 }
 
 // static
+PyDoc_STRVAR(listReserved_doc,
+    "lst.reserved() -> number of elements lst could hold, as currently allocated\n"
+    "\n"
+    "Contrast this with the actual length len(lst) of the lst.\n"
+    "lst.reserved() >= len(lst)\n"
+    );
 PyObject* PyListOfInstance::listReserved(PyObject* o, PyObject* args) {
     if (PyTuple_Size(args) != 0) {
         PyErr_SetString(PyExc_TypeError, "ListOf.reserved takes no arguments");
@@ -604,16 +802,24 @@ PyObject* PyListOfInstance::listReserved(PyObject* o, PyObject* args) {
     return PyLong_FromLong(self_w->type()->reserved(self_w->dataPtr()));
 }
 
+PyDoc_STRVAR(listReserve_doc,
+    "lst.reserve(n) -> None, and allocates space to potentially hold n elements\n"
+    "\n"
+    "Expands or shrinks the memory allocated for lst.\n"
+    "Won't shrink lst smaller than the elements it currently contains.\n"
+    "For example, n=0 will reallocate lst to the minimum possible size, \n"
+    "containing len(lst) elements and no unused allocation.\n"
+    );
 PyObject* PyListOfInstance::listReserve(PyObject* o, PyObject* args) {
     if (PyTuple_Size(args) != 1) {
-        PyErr_SetString(PyExc_TypeError, "ListOf.append takes one argument");
+        PyErr_SetString(PyExc_TypeError, "ListOf.reserve takes one argument");
         return NULL;
     }
 
     PyObject* pyReserveSize = PyTuple_GetItem(args, 0);
 
     if (!PyLong_Check(pyReserveSize)) {
-        PyErr_SetString(PyExc_TypeError, "ListOf.append takes an integer");
+        PyErr_SetString(PyExc_TypeError, "ListOf.reserve takes an integer");
         return NULL;
     }
 
@@ -626,6 +832,9 @@ PyObject* PyListOfInstance::listReserve(PyObject* o, PyObject* args) {
     return incref(Py_None);
 }
 
+PyDoc_STRVAR(listClear_doc,
+    "lst.clear() -> None, and resize lst to 0."
+    );
 PyObject* PyListOfInstance::listClear(PyObject* o, PyObject* args) {
     if (PyTuple_Size(args) != 0) {
         PyErr_SetString(PyExc_TypeError, "ListOf.clear takes no arguments");
@@ -639,17 +848,24 @@ PyObject* PyListOfInstance::listClear(PyObject* o, PyObject* args) {
     return incref(Py_None);
 }
 
+PyDoc_STRVAR(listResize_doc,
+    "lst.resize(n[, e]) -> None, and lst now contains n elements\n"
+    "\n"
+    "if n > len(lst), default-constructed elements (or e if provided) are\n"
+    "placed the end of lst.\n"
+    "if n < len(lst), elements are deleted from the end of lst.\n"
+    );
 PyObject* PyListOfInstance::listResize(PyObject* o, PyObject* args) {
     try {
         if (PyTuple_Size(args) != 1 && PyTuple_Size(args) != 2) {
-            PyErr_SetString(PyExc_TypeError, "ListOf.append takes one argument");
+            PyErr_SetString(PyExc_TypeError, "ListOf.resize takes one argument");
             return NULL;
         }
 
         PyObject* pySize = PyTuple_GetItem(args, 0);
 
         if (!PyLong_Check(pySize)) {
-            PyErr_SetString(PyExc_TypeError, "ListOf.append takes an integer");
+            PyErr_SetString(PyExc_TypeError, "ListOf.resize takes an integer");
             return NULL;
         }
 
@@ -663,14 +879,19 @@ PyObject* PyListOfInstance::listResize(PyObject* o, PyObject* args) {
         } else {
             if (PyTuple_Size(args) == 2) {
                 Instance temp(eltType, [&](instance_ptr data) {
-                    PyInstance::copyConstructFromPythonInstance(eltType, data, PyTuple_GetItem(args, 1), true);
+                    PyInstance::copyConstructFromPythonInstance(
+                        eltType,
+                        data,
+                        PyTuple_GetItem(args, 1),
+                        ConversionLevel::ImplicitContainers
+                    );
                 });
 
                 self_w->type()->resize(self_w->dataPtr(), size, temp.data());
             } else {
                 if (!self_w->type()->getEltType()->is_default_constructible()) {
                     PyErr_SetString(PyExc_TypeError, "Cannot increase the size of this list without an object to copy in because the"
-                        " element type is not copy-constructible");
+                        " element type is not default-constructible");
                     return NULL;
                 }
 
@@ -687,6 +908,14 @@ PyObject* PyListOfInstance::listResize(PyObject* o, PyObject* args) {
     }
 }
 
+PyDoc_STRVAR(listPop_doc,
+    "lst.pop() -> last element of lst, and remove last element from lst\n"
+    "lst.pop(n) -> nth element of lst, and remove nth element from lst\n"
+    "\n"
+    "Raises IndexError when called on an empty list.\n"
+    "Raises IndexError when n is out of range.\n"
+    "Negative n is interpreted as in slice notation.\n"
+    );
 PyObject* PyListOfInstance::listPop(PyObject* o, PyObject* args) {
     if (PyTuple_Size(args) != 0 && PyTuple_Size(args) != 1) {
         PyErr_SetString(PyExc_TypeError, "ListOf.pop takes zero or one argument");
@@ -699,7 +928,7 @@ PyObject* PyListOfInstance::listPop(PyObject* o, PyObject* args) {
         PyObject* pySize = PyTuple_GetItem(args, 0);
 
         if (!PyLong_Check(pySize)) {
-            PyErr_SetString(PyExc_TypeError, "ListOf.append takes an integer");
+            PyErr_SetString(PyExc_TypeError, "ListOf.pop takes an integer");
             return NULL;
         }
 
@@ -711,7 +940,7 @@ PyObject* PyListOfInstance::listPop(PyObject* o, PyObject* args) {
     int64_t listSize = self_w->type()->count(self_w->dataPtr());
 
     if (listSize == 0) {
-        PyErr_SetString(PyExc_TypeError, "pop from empty list");
+        PyErr_SetString(PyExc_IndexError, "pop from empty list");
         return NULL;
     }
 
@@ -734,9 +963,18 @@ PyObject* PyListOfInstance::listPop(PyObject* o, PyObject* args) {
     return result;
 }
 
+PyDoc_STRVAR(listTranspose_doc,
+    "lst.transpose() -> transposition of ListOf(Tuple()) or ListOf(NamedTuple())\n"
+    "\n"
+    "This is a matrix transposition that swaps the types of the rows and columns.\n"
+    "If lst is a ListOf(Tuple(X1, X2, X3)), then lst.transpose() is a\n"
+    "Tuple(ListOf(X1), ListOf(X2), ListOf(X3)).\n"
+    "If lst is a ListOf(NamedTuple(a=X1, b=X2, c=X3)), then lst.transpose() is a\n"
+    "NamedTuple(a=ListOf(X1), b=ListOf(X2), c=ListOf(X3)).\n"
+    );
 PyObject* PyListOfInstance::listTranspose(PyObject* o, PyObject* args) {
     if (PyTuple_Size(args) != 0) {
-        PyErr_SetString(PyExc_TypeError, "ListOf.trasnpose takes zero arguments");
+        PyErr_SetString(PyExc_TypeError, "ListOf.transpose takes zero arguments");
         return NULL;
     }
 
@@ -812,8 +1050,18 @@ int PyListOfInstance::mp_ass_subscript_concrete(PyObject* item, PyObject* value)
 
     Type* eltType = type()->getEltType();
 
-    if (PyLong_Check(item)) {
-        int64_t ix = PyLong_AsLongLong(item);
+    PyObjectHolder index(item);
+
+    if (!PyLong_Check(item)) {
+        index.set(PyNumber_Index(item));
+
+        if (!index) {
+            throw PythonExceptionSet();
+        }
+    }
+
+    if (PyLong_Check(index)) {
+        int64_t ix = PyLong_AsLongLong(index);
         int64_t count = type()->count(dataPtr());
 
         if (ix < 0) {
@@ -832,9 +1080,16 @@ int PyListOfInstance::mp_ass_subscript_concrete(PyObject* item, PyObject* value)
                 type()->eltPtr(dataPtr(), ix),
                 value_w->dataPtr()
                 );
+
+            return 0;
         } else {
             Instance toAssign(eltType, [&](instance_ptr data) {
-                PyInstance::copyConstructFromPythonInstance(eltType, data, value, true);
+                PyInstance::copyConstructFromPythonInstance(
+                    eltType,
+                    data,
+                    value,
+                    ConversionLevel::ImplicitContainers
+                );
             });
 
             eltType->assign(
@@ -846,32 +1101,50 @@ int PyListOfInstance::mp_ass_subscript_concrete(PyObject* item, PyObject* value)
         }
     }
 
-    return ((PyInstance*)this)->mp_ass_subscript_concrete(item, value);
+    return PyInstance::mp_ass_subscript_concrete(item, value);
 }
 
+PyDoc_STRVAR(
+    LIST_TO_BYTES_DOCSTRING,
+    "ListOf(T).toBytes(self) -> bytes\n\nConvert a ListOf(T) to the raw bytes underneath it.\n\n"
+    "This is only valid if T is \"POD\" (Plain Old Data), meaning it must be\n"
+    "an integer, float, or bool type, or some combination of those in a Tuple or \n"
+    "NamedTuple."
+);
+
+PyDoc_STRVAR(
+    LIST_FROM_BYTES_DOCSTRING,
+    "ListOf(T).fromBytes(b: bytes) -> ListOf(T)\n\nConstruct a ListOf(T) from the raw bytes that would underly it.\n\n"
+    "This is only valid if T is \"POD\" (Plain Old Data), meaning it must be\n"
+    "an integer, float, or bool type, or some combination of those in a Tuple or \n"
+    "NamedTuple."
+);
+
 PyMethodDef* PyListOfInstance::typeMethodsConcrete(Type* t) {
-    return new PyMethodDef [12] {
-        {"toArray", (PyCFunction)PyTupleOrListOfInstance::toArray, METH_VARARGS, NULL},
-        {"append", (PyCFunction)PyListOfInstance::listAppend, METH_VARARGS, NULL},
-        {"extend", (PyCFunction)PyListOfInstance::listExtend, METH_VARARGS, NULL},
-        {"clear", (PyCFunction)PyListOfInstance::listClear, METH_VARARGS, NULL},
-        {"reserved", (PyCFunction)PyListOfInstance::listReserved, METH_VARARGS, NULL},
-        {"reserve", (PyCFunction)PyListOfInstance::listReserve, METH_VARARGS, NULL},
-        {"resize", (PyCFunction)PyListOfInstance::listResize, METH_VARARGS, NULL},
-        {"pop", (PyCFunction)PyListOfInstance::listPop, METH_VARARGS, NULL},
-        {"setSizeUnsafe", (PyCFunction)PyListOfInstance::listSetSizeUnsafe, METH_VARARGS, NULL},
-        {"pointerUnsafe", (PyCFunction)PyListOfInstance::listPointerUnsafe, METH_VARARGS, NULL},
-        {"transpose", (PyCFunction)PyListOfInstance::listTranspose, METH_VARARGS, NULL},
+    return new PyMethodDef [14] {
+        {"toArray", (PyCFunction)PyTupleOrListOfInstance::toArray, METH_VARARGS, ListOf_toArray_doc},
+        {"toBytes", (PyCFunction)PyTupleOrListOfInstance::toBytes, METH_VARARGS, LIST_TO_BYTES_DOCSTRING},
+        {"fromBytes", (PyCFunction)PyTupleOrListOfInstance::fromBytes, METH_VARARGS | METH_KEYWORDS | METH_CLASS, LIST_FROM_BYTES_DOCSTRING},
+        {"append", (PyCFunction)PyListOfInstance::listAppend, METH_VARARGS, listAppend_doc},
+        {"extend", (PyCFunction)PyListOfInstance::listExtend, METH_VARARGS, listExtend_doc},
+        {"clear", (PyCFunction)PyListOfInstance::listClear, METH_VARARGS, listClear_doc},
+        {"reserved", (PyCFunction)PyListOfInstance::listReserved, METH_VARARGS, listReserved_doc},
+        {"reserve", (PyCFunction)PyListOfInstance::listReserve, METH_VARARGS, listReserve_doc},
+        {"resize", (PyCFunction)PyListOfInstance::listResize, METH_VARARGS, listResize_doc},
+        {"pop", (PyCFunction)PyListOfInstance::listPop, METH_VARARGS, listPop_doc},
+        {"setSizeUnsafe", (PyCFunction)PyListOfInstance::listSetSizeUnsafe, METH_VARARGS, listSetSizeUnsafe_doc},
+        {"pointerUnsafe", (PyCFunction)PyTupleOrListOfInstance::pointerUnsafe, METH_VARARGS, listPointerUnsafe_doc},
+        {"transpose", (PyCFunction)PyListOfInstance::listTranspose, METH_VARARGS, listTranspose_doc},
         {NULL, NULL}
     };
 }
 
 void PyListOfInstance::constructFromPythonArgumentsConcrete(ListOfType* t, uint8_t* data, PyObject* args, PyObject* kwargs) {
-    if (PyTuple_Size(args) == 1 && !kwargs) {
+    if ((!kwargs || PyDict_Size(kwargs) == 0) && (args && PyTuple_Size(args) == 1)) {
         PyObject* arg = PyTuple_GetItem(args, 0);
         Type* argType = extractTypeFrom(arg->ob_type);
 
-        if (argType == t) {
+        if (Type::typesEquivalent(argType, t)) {
             //following python semantics, this needs to produce a new object
             //that's a copy of the original list. We can't just incref it and return
             //the original object because it has state.
@@ -897,3 +1170,127 @@ int PyTupleOrListOfInstance::pyInquiryConcrete(const char* op, const char* opErr
     // op == '__bool__'
     return type()->count(dataPtr()) != 0;
 }
+
+bool PyListOfInstance::compare_to_python_concrete(ListOfType* listT, instance_ptr self, PyObject* other, bool exact, int pyComparisonOp) {
+    auto convert = [&](char cmpValue) { return cmpResultToBoolForPyOrdering(pyComparisonOp, cmpValue); };
+
+    Type* otherType = extractTypeFrom(other->ob_type);
+
+    if (PyList_Check(other) || (otherType && otherType->getTypeCategory() == Type::TypeCategory::catListOf)) {
+        int lenS = listT->count(self);
+        int indexInOwn = 0;
+
+        int result = 0;
+
+        iterateWithEarlyExit(other, [&](PyObject* listItem) {
+            if (indexInOwn >= lenS) {
+                // we ran out of items in our list
+                result = -1;
+                return false;
+            }
+
+            if (!compare_to_python(listT->getEltType(), listT->eltPtr(self, indexInOwn), listItem, exact, Py_EQ)) {
+                if (pyComparisonOp == Py_EQ || pyComparisonOp == Py_NE) {
+                    result = 1;
+                    return false;
+                }
+                if (compare_to_python(listT->getEltType(), listT->eltPtr(self, indexInOwn), listItem, exact, Py_LT)) {
+                    result = -1;
+                    return false;
+                }
+
+                result = 1;
+                return false;
+            }
+
+            indexInOwn += 1;
+
+            return true;
+        });
+
+        if (result) {
+            return convert(result);
+        }
+
+        if (indexInOwn == lenS) {
+            return convert(0);
+        }
+
+        return convert(1);
+    }
+
+    if (pyComparisonOp == Py_EQ || pyComparisonOp == Py_NE) {
+        return convert(1);
+    }
+
+    PyErr_Format(
+        PyExc_TypeError,
+        "Comparison not supported between instances of '%s' and '%s'.",
+        listT->name().c_str(),
+        other->ob_type->tp_name
+    );
+
+    throw PythonExceptionSet();
+}
+
+bool PyTupleOfInstance::compare_to_python_concrete(TupleOfType* tupT, instance_ptr self, PyObject* other, bool exact, int pyComparisonOp) {
+    auto convert = [&](char cmpValue) { return cmpResultToBoolForPyOrdering(pyComparisonOp, cmpValue); };
+
+    Type* otherType = extractTypeFrom(other->ob_type);
+
+    if (PyTuple_Check(other) || (otherType && (otherType->getTypeCategory() == Type::TypeCategory::catTupleOf || otherType->isComposite()))) {
+        int lenS = tupT->count(self);
+        int indexInOwn = 0;
+
+        int result = 0;
+
+        iterateWithEarlyExit(other, [&](PyObject* tupleItem) {
+            if (indexInOwn >= lenS) {
+                // we ran out of items in our list
+                result = -1;
+                return false;
+            }
+
+            if (!compare_to_python(tupT->getEltType(), tupT->eltPtr(self, indexInOwn), tupleItem, exact, Py_EQ)) {
+                if (pyComparisonOp == Py_EQ || pyComparisonOp == Py_NE) {
+                    result = 1;
+                    return false;
+                }
+                if (compare_to_python(tupT->getEltType(), tupT->eltPtr(self, indexInOwn), tupleItem, exact, Py_LT)) {
+                    result = -1;
+                    return false;
+                }
+
+                result = 1;
+                return false;
+            }
+
+            indexInOwn += 1;
+
+            return true;
+        });
+
+        if (result) {
+            return convert(result);
+        }
+
+        if (indexInOwn == lenS) {
+            return convert(0);
+        }
+
+        return convert(1);
+    }
+
+    if (pyComparisonOp == Py_EQ || pyComparisonOp == Py_NE) {
+        return convert(1);
+    }
+
+    PyErr_Format(
+        PyExc_TypeError,
+        "Comparison not supported between instances of '%s' and '%s'.",
+        tupT->name().c_str(),
+        other->ob_type->tp_name
+    );
+
+    throw PythonExceptionSet();
+}
