@@ -1,4 +1,4 @@
-#   Copyright 2017-2019 typed_python Authors
+#   Copyright 2017-2020 typed_python Authors
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
@@ -75,6 +75,7 @@ Type = Type.define(Alternative(
         Constant.Int(val=0, bits=self.bits, signed=self.signed) if self.matches.Int else
         Constant.Struct(elements=[(name, t.zero()) for name, t in self.element_types]) if self.matches.Struct else
         Constant.NullPointer(value_type=self.value_type) if self.matches.Pointer else
+        Constant.Array(value_type=self, values=[self.element_type.zero()] * self.count) if self.matches.Array else
         raising(Exception("Can't make a zero value from %s" % self))
     )
 ))
@@ -83,6 +84,8 @@ Type = Type.define(Alternative(
 def const_truth_value(c):
     if c.matches.Int:
         return c.val != 0
+    if c.matches.Float:
+        return c.val != 0
     return False
 
 
@@ -105,6 +108,8 @@ def const_str(c):
         return "nullptr"
     if c.matches.Void:
         return "void"
+    if c.matches.Array:
+        return "Array[" + ",".join(str(x) for x in c.values) + "]"
 
     assert False, type(c)
 
@@ -117,6 +122,7 @@ Constant = Constant.define(Alternative(
     Int={'val': int, 'bits': int, 'signed': bool},
     Struct={'elements': TupleOf(Tuple(str, Constant))},
     ByteArray={'val': bytes},
+    Array={'value_type': Type, 'values': TupleOf(Constant)},
     NullPointer={'value_type': Type},
     truth_value=const_truth_value,
     __str__=const_str
@@ -273,8 +279,6 @@ def expr_str(self):
         return "(%s(%s))" % (str(self.op), str(self.operand))
     if self.matches.Variable:
         return self.name
-    if self.matches.Attribute:
-        return "(" + str(self.left) + ")." + self.attr
     if self.matches.ElementPtr:
         return "(" + str(self.left) + ").gep" + \
             "(" + ",".join(str(r) for r in self.offsets) + ")"
@@ -315,9 +319,14 @@ def expr_str(self):
             return "while " + c + ":\n"\
                    + indent(t).rstrip() + "\nelse:\n" + indent(f).rstrip()
     if self.matches.Return:
+        if self.blockName is not None:
+            return "return to " + self.blockName
+
         s = (str(self.arg) if self.arg is not None else "")
+
         if "\n" in s:
             return "return (" + s + ")"
+
         return "return " + s
     if self.matches.Let:
         if self.val.matches.Sequence and len(self.val.vals) > 1:
@@ -341,8 +350,9 @@ def expr_str(self):
     if self.matches.Sequence:
         return "\n".join(str(x) for x in self.vals)
     if self.matches.Finally:
+        label = " [label=" + self.name + "]" if self.name else ""
         return (
-            "try:\n" + indent(str(self.expr)) + "\nfinally:\n"
+            "try:\n" + indent(str(self.expr)) + "\nfinally " + label + ":\n"
             + indent("\n".join(str(x) for x in self.teardowns))
         )
     if self.matches.TryCatch:
@@ -365,8 +375,15 @@ def expr_str(self):
         return "mark slot %s initialized" % self.name
     if self.matches.StackSlot:
         return "slot(name=%s,t=%s)" % (self.name, str(self.type))
+    if self.matches.GlobalVariable:
+        return "global(name=%s,t=%s)" % (self.name, str(self.type))
+    if self.matches.ExceptionPropagator:
+        return (
+            "try:\n" + indent(str(self.expr)) + "\n" +
+            "rethrow with %s:\n" % self.varname + indent(str(self.handler)).rstrip()
+        )
 
-    assert False
+    assert False, type(self)
 
 
 def expr_is_simple(expr):
@@ -385,6 +402,69 @@ def expr_is_simple(expr):
     return False
 
 
+def expr_return_targets(expr):
+    if expr.matches.Finally:
+        return expr.expr.returnTargets() - set([expr.name])
+
+    if expr.matches.Return:
+        if expr.blockName:
+            return set([expr.blockName])
+        else:
+            return set()
+
+    if expr.matches.MakeStruct:
+        res = set()
+        for _, e in expr.args:
+            res |= e.returnTargets()
+        return res
+
+    res = set()
+
+    for name in expr.ElementType.ElementNames:
+        child = getattr(expr, name)
+
+        if isinstance(child, Expression):
+            res |= child.returnTargets()
+        elif isinstance(child, TupleOf(Expression)):
+            for c in child:
+                res |= c.returnTargets()
+
+    return res
+
+
+def expr_with_return_target_name(self, name):
+    return Expression.Finally(
+        expr=self,
+        name=name
+    )
+
+
+def expr_could_throw(self):
+    if self.matches.Throw:
+        return True
+
+    if self.matches.Call:
+        return True
+
+    if self.matches.MakeStruct:
+        for _, e in self.args:
+            if e.couldThrow():
+                return True
+
+    for name in self.ElementType.ElementNames:
+        child = getattr(self, name)
+
+        if isinstance(child, Expression):
+            if child.couldThrow():
+                return True
+        elif isinstance(child, TupleOf(Expression)):
+            for c in child:
+                if c.couldThrow():
+                    return True
+
+    return False
+
+
 Expression = Expression.define(Alternative(
     "Expression",
     Constant={'val': Constant},
@@ -409,24 +489,38 @@ Expression = Expression.define(Alternative(
         'false': Expression
     },
     Throw={'expr': Expression },  # throw a pointer.
+    # evaluate 'expr', which must have type 'void' if it returns. if it throws an exception,
+    # evaluate 'handler', which must also have type 'void' if it returns, with the exception
+    # bound to 'varname'
     TryCatch={
         'expr': Expression,
         'varname': str,  # varname is bound to a int8*
         'handler': Expression
     },
+    # evaluate 'expr', which can have any type. If it throws, evaluate
+    # 'handler', which must propagate the exception upward.
+    ExceptionPropagator={
+        'expr': Expression,
+        'varname': str,  # varname is bound to a int8*
+        'handler': Expression
+    },
     While={
         'cond': Expression,
         'while_true': Expression,
         'orelse': Expression
     },
-    Return={'arg': OneOf(Expression, None)},
+    # return control flow to a higher point in the stack. If 'name' is None, exit the function.
+    # otherwise, search for the first 'finally' block above it with that name and return to that
+    # scope. In that case 'arg' must be 'None'.
+    Return={'arg': OneOf(None, Expression), 'blockName': OneOf(None, str)},
     Let={'var': str, 'val': Expression, 'within': Expression},
     # evaluate 'expr', and then call teardowns if we passed through a named 'ActivatesTeardown'
-    # clause
-    Finally={'expr': Expression, 'teardowns': TupleOf(Teardown)},
+    # clause. If name is nonempty, then we can explicitly 'return' to it
+    Finally={'expr': Expression, 'teardowns': TupleOf(Teardown), 'name': OneOf(None, str)},
     Sequence={'vals': TupleOf(Expression)},
     ActivatesTeardown={'name': str},
     StackSlot={'name': str, 'type': Type},
+    GlobalVariable={'name': str, 'type': Type, 'metadata': object},
     ElementPtrIntegers=lambda self, *offsets:
         Expression.ElementPtr(
             left=self,
@@ -458,13 +552,17 @@ Expression = Expression.define(Alternative(
     rshift=lambda self, other: Expression.Binop(op=BinaryOp.RShift(), left=self, right=ensureExpr(other)),
     bitand=lambda self, other: Expression.Binop(op=BinaryOp.BitAnd(), left=self, right=ensureExpr(other)),
     bitor=lambda self, other: Expression.Binop(op=BinaryOp.BitOr(), left=self, right=ensureExpr(other)),
+    bitxor=lambda self, other: Expression.Binop(op=BinaryOp.BitXor(), left=self, right=ensureExpr(other)),
     load=lambda self: Expression.Load(ptr=self),
     store=lambda self, val: Expression.Store(ptr=self, val=ensureExpr(val)),
     atomic_add=lambda self, val: Expression.AtomicAdd(ptr=self, val=ensureExpr(val)),
     cast=lambda self, targetType: Expression.Cast(left=self, to_type=targetType),
     with_comment=lambda self, c: Expression.Comment(comment=c, expr=self),
     elemPtr=lambda self, *exprs: Expression.ElementPtr(left=self, offsets=[ensureExpr(e) for e in exprs]),
-    is_simple=expr_is_simple
+    is_simple=expr_is_simple,
+    returnTargets=expr_return_targets,
+    withReturnTargetName=expr_with_return_target_name,
+    couldThrow=expr_could_throw
 ))
 
 
@@ -492,6 +590,12 @@ def const_float_expr(f):
     )
 
 
+def const_float32_expr(f):
+    return Expression.Constant(
+        val=Constant.Float(bits=32, val=f)
+    )
+
+
 def const_int_expr(i):
     return Expression.Constant(
         val=Constant.Int(bits=64, val=i, signed=True)
@@ -504,9 +608,15 @@ def const_uint64_expr(i):
     )
 
 
-def const_int32_expr(i):
+def const_uint32_expr(i):
     return Expression.Constant(
-        val=Constant.Int(bits=32, val=i, signed=True)
+        val=Constant.Int(bits=32, val=i, signed=False)
+    )
+
+
+def const_uint16_expr(i):
+    return Expression.Constant(
+        val=Constant.Int(bits=16, val=i, signed=False)
     )
 
 
@@ -516,6 +626,24 @@ def const_uint8_expr(i):
     )
 
 
+def const_int32_expr(i):
+    return Expression.Constant(
+        val=Constant.Int(bits=32, val=i, signed=True)
+    )
+
+
+def const_int16_expr(i):
+    return Expression.Constant(
+        val=Constant.Int(bits=16, val=i, signed=True)
+    )
+
+
+def const_int8_expr(i):
+    return Expression.Constant(
+        val=Constant.Int(bits=8, val=i, signed=True)
+    )
+
+
 def const_bool_expr(i):
     return Expression.Constant(
         val=Constant.Int(bits=1, val=i, signed=False)
