@@ -13,8 +13,9 @@
 #   limitations under the License.
 
 import typed_python.python_ast as python_ast
-from typed_python.type_promotion import computeArithmeticBinaryResultType
-
+from typed_python.compiler.conversion_level import ConversionLevel
+from typed_python.type_promotion import computeArithmeticBinaryResultType, bitness, signedness, floatness, isSignedInt
+from typed_python import _types
 import typed_python.compiler.type_wrappers.runtime_functions as runtime_functions
 from typed_python.compiler.type_wrappers.wrapper import Wrapper
 import typed_python.compiler.native_ast as native_ast
@@ -22,7 +23,7 @@ import typed_python.compiler
 from math import trunc, floor, ceil
 
 from typed_python import (
-    Float32, Float64, Int64, Bool, String, Int8, UInt8, Int16, UInt16, Int32, UInt32, UInt64
+    Float32, Int8, UInt8, Int16, UInt16, Int32, UInt32, UInt64
 )
 
 pyOpToNative = {
@@ -87,6 +88,15 @@ class ArithmeticTypeWrapper(Wrapper):
     def convert_destroy(self, context, instance):
         pass
 
+    def convert_index_cast(self, context, expr):
+        if expr.expr_type.typeRepresentation is float:
+            return super().convert_index_cast(context, expr)
+
+        return expr.toInt64()
+
+    def convert_math_float_cast(self, context, e):
+        return e.toFloat64()
+
     def convert_unary_op(self, context, instance, op):
         if op.matches.USub:
             return context.pushPod(self, instance.nonref_expr.negate())
@@ -96,29 +106,59 @@ class ArithmeticTypeWrapper(Wrapper):
 
         return super().convert_unary_op(context, instance, op)
 
-    def _can_convert_to_type(self, otherType, explicit):
-        if not explicit:
-            return self == otherType
+    def _can_convert_to_type(self, otherType, conversionLevel):
+        if self == otherType:
+            return True
 
-        return isinstance(otherType, ArithmeticTypeWrapper)
+        if conversionLevel.isNewOrHigher() and otherType.typeRepresentation is str:
+            return True
 
-    def _can_convert_from_type(self, otherType, explicit):
+        if not isinstance(otherType, ArithmeticTypeWrapper):
+            return False
+
+        if not isValidConversion(
+            self.typeRepresentation,
+            otherType.typeRepresentation,
+            conversionLevel
+        ):
+            return False
+
+        if isinstance(self, FloatWrapper) and isinstance(otherType, IntWrapper) and otherType.typeRepresentation is not bool:
+            return "Maybe"
+
+        return True
+
+    def _can_convert_from_type(self, otherType, conversionLevel):
         return False
 
+    def convert_to_type_constant(self, context, expr, target_type, level: ConversionLevel):
+        """Given that 'expr' is a constant expression, attempt to convert it directly.
+
+        This function should return None if it can't convert it to a constant, otherwise
+        a typed expression with the constant.
+        """
+        if target_type.typeRepresentation in (str, int, float, bool, UInt64, UInt32, UInt16, UInt8, Int8, Int16, Int32):
+            if isValidConversion(target_type.typeRepresentation, self.typeRepresentation, level):
+                try:
+                    return context.constant(target_type.typeRepresentation(expr.constantValue))
+                except Exception as e:
+                    context.pushException(type(e), *e.args)
+                    return "FAILURE"
+
     def convert_type_call(self, context, typeInst, args, kwargs):
         if len(args) == 0 and not kwargs:
             return context.push(self, lambda x: x.convert_default_initialize())
 
         if len(args) == 1 and not kwargs:
-            return args[0].convert_cast(self)
+            return args[0].convert_to_type(self, ConversionLevel.New)
 
         return super().convert_type_call(context, typeInst, args, kwargs)
 
 
 def toWrapper(T):
-    if T is Bool:
+    if T is bool:
         return BoolWrapper()
-    if T.IsInteger:
+    if not floatness(T):
         return IntWrapper(T)
     return FloatWrapper(T)
 
@@ -126,13 +166,20 @@ def toWrapper(T):
 def toFloatType(T1):
     """Convert an int or float type to the enclosing float type."""
     if not T1.IsFloat:
-        if T1.Bits <= 32:
+        if bitness(T1) <= 32:
             return Float32
         else:
-            return Float64
+            return float
     return T1
 
 
+def isValidConversion(fromType, toType, conversionLevel):
+    if (fromType is str or toType is str) and conversionLevel >= ConversionLevel.New:
+        return True
+
+    return _types.isValidArithmeticConversion(fromType, toType, conversionLevel.LEVEL)
+
+
 class IntWrapper(ArithmeticTypeWrapper):
     def __init__(self, T):
         super().__init__(T)
@@ -140,72 +187,77 @@ class IntWrapper(ArithmeticTypeWrapper):
     def getNativeLayoutType(self):
         T = self.typeRepresentation
 
-        return native_ast.Type.Int(bits=T.Bits, signed=T.IsSignedInt)
+        return native_ast.Type.Int(bits=bitness(T), signed=signedness(T))
 
     def convert_hash(self, context, expr):
-        if self.typeRepresentation == Int64:
+        if self.typeRepresentation == int:
             return context.pushPod(Int32, runtime_functions.hash_int64.call(expr.nonref_expr))
 
         if self.typeRepresentation == UInt64:
             return context.pushPod(Int32, runtime_functions.hash_uint64.call(expr.nonref_expr))
 
-        return expr.convert_to_type(Int32)
+        return expr.convert_to_type(Int32, ConversionLevel.Implicit)
 
-    def convert_to_type_with_target(self, context, e, targetVal, explicit):
+    def convert_to_type_with_target(self, context, instance, targetVal, conversionLevel, mayThrowOnFailure=False):
         assert targetVal.isReference
 
         target_type = targetVal.expr_type
 
-        if not explicit:
-            return super().convert_to_type_with_target(context, e, targetVal, explicit)
+        if not isValidConversion(
+            self.typeRepresentation,
+            targetVal.expr_type.typeRepresentation,
+            conversionLevel
+        ):
+            return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
         if isinstance(target_type, FloatWrapper):
             context.pushEffect(
                 targetVal.expr.store(
                     native_ast.Expression.Cast(
-                        left=e.nonref_expr,
-                        to_type=native_ast.Type.Float(bits=target_type.typeRepresentation.Bits)
+                        left=instance.nonref_expr,
+                        to_type=native_ast.Type.Float(bits=bitness(target_type.typeRepresentation))
                     )
                 )
             )
             return context.constant(True)
 
-        if target_type.typeRepresentation == Bool and explicit:
-            targetVal.convert_copy_initialize(e != 0)
-            return context.constant(True)
-
         if isinstance(target_type, IntWrapper):
             context.pushEffect(
                 targetVal.expr.store(
                     native_ast.Expression.Cast(
-                        left=e.nonref_expr,
+                        left=instance.nonref_expr,
                         to_type=native_ast.Type.Int(
-                            bits=target_type.typeRepresentation.Bits,
-                            signed=target_type.typeRepresentation.IsSignedInt
+                            bits=bitness(target_type.typeRepresentation),
+                            signed=signedness(target_type.typeRepresentation)
                         )
                     )
                 )
             )
             return context.constant(True)
 
-        return super().convert_to_type_with_target(context, e, targetVal, explicit)
+        if isinstance(target_type, BoolWrapper):
+            context.pushEffect(
+                targetVal.expr.store(
+                    instance.nonref_expr.neq(instance.expr_type.getNativeLayoutType().zero())
+                )
+            )
+            return context.constant(True)
 
-    def convert_cast(self, context, instance, target_type):
-        if target_type.typeRepresentation == String:
-            if self.typeRepresentation == Int64:
-                return context.push(
-                    str,
-                    lambda strRef: strRef.expr.store(
-                        runtime_functions.int64_to_string.call(instance.nonref_expr).cast(strRef.expr_type.layoutType)
+        if target_type.typeRepresentation is str:
+            if self.typeRepresentation == int:
+                context.pushEffect(
+                    targetVal.expr.store(
+                        runtime_functions.int64_to_string.call(instance.nonref_expr).cast(targetVal.expr_type.layoutType)
                     )
                 )
+                return context.constant(True)
             elif self.typeRepresentation == UInt64:
-                return context.push(
-                    str,
-                    lambda strRef: strRef.expr.store(
-                        runtime_functions.uint64_to_string.call(instance.nonref_expr).cast(strRef.expr_type.layoutType)
+                context.pushEffect(
+                    targetVal.expr.store(
+                        runtime_functions.uint64_to_string.call(instance.nonref_expr).cast(targetVal.expr_type.layoutType)
                     )
                 )
+                return context.constant(True)
             else:
                 suffix = {
                     Int32: 'i32',
@@ -216,12 +268,18 @@ class IntWrapper(ArithmeticTypeWrapper):
                     UInt8: 'u8'
                 }[self.typeRepresentation]
 
-                return instance.convert_to_type(int).convert_cast(str) + context.constant(suffix)
+                targetVal.convert_copy_initialize(
+                    instance.convert_to_type(
+                        int,
+                        ConversionLevel.Implicit
+                    ).convert_to_type(str, ConversionLevel.New) + context.constant(suffix)
+                )
+                return context.constant(True)
 
-        return super().convert_cast(context, instance, target_type)
+        return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
     def convert_abs(self, context, expr):
-        if self.typeRepresentation.IsSignedInt:
+        if isSignedInt(self.typeRepresentation):
             return context.pushPod(
                 self,
                 native_ast.Expression.Branch(
@@ -234,17 +292,30 @@ class IntWrapper(ArithmeticTypeWrapper):
             return context.pushPod(self, expr.nonref_expr)
 
     def convert_builtin(self, f, context, expr, a1=None):
+        if f is chr and a1 is None:
+            if expr.isConstant:
+                return context.constant(chr(expr.constantValue))
+
+            return context.push(
+                str,
+                lambda strRef: strRef.expr.store(
+                    runtime_functions.string_chr_int64.call(
+                        expr.toInt64().nonref_expr
+                    ).cast(strRef.expr_type.layoutType)
+                )
+            )
+
         if f is round:
             if a1 is None:
                 return context.pushPod(
                     float,
                     runtime_functions.round_float64.call(expr.toFloat64().nonref_expr, context.constant(0))
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
             else:
                 return context.pushPod(
                     float,
                     runtime_functions.round_float64.call(expr.toFloat64().nonref_expr, a1.toInt64().nonref_expr)
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
 
         if f in [trunc, floor, ceil]:
             return context.pushPod(self, expr.nonref_expr)
@@ -253,18 +324,26 @@ class IntWrapper(ArithmeticTypeWrapper):
 
     def convert_unary_op(self, context, left, op):
         if op.matches.Not:
-            return context.pushPod(self, left.nonref_expr.logical_not())
+            if left.isConstant:
+                return context.constant(not left.constantValue)
+            return context.pushPod(bool, left.nonref_expr.logical_not())
         if op.matches.Invert:
+            if left.isConstant:
+                return context.constant(~left.constantValue)
             return context.pushPod(self, left.nonref_expr.bitwise_not())
         if op.matches.USub:
+            if left.isConstant:
+                return context.constant(-left.constantValue)
             return context.pushPod(self, left.nonref_expr.negate())
         if op.matches.UAdd:
+            if left.isConstant:
+                return context.constant(left.constantValue)
             return context.pushPod(self, left.nonref_expr)
 
         return super().convert_unary_op(context, left, op)
 
     def convert_bin_op(self, context, left, op, right, inplace):
-        if op.matches.Div:
+        if op.matches.Div and isinstance(right.expr_type, ArithmeticTypeWrapper):
             T = toWrapper(
                 computeArithmeticBinaryResultType(
                     computeArithmeticBinaryResultType(
@@ -274,7 +353,13 @@ class IntWrapper(ArithmeticTypeWrapper):
                     Float32
                 )
             )
-            return left.convert_to_type(T).convert_bin_op(op, right.convert_to_type(T))
+            return left.convert_to_type(
+                T,
+                ConversionLevel.Implicit
+            ).convert_bin_op(
+                op,
+                right.convert_to_type(T, ConversionLevel.Implicit)
+            )
 
         if right.expr_type != self:
             if isinstance(right.expr_type, ArithmeticTypeWrapper):
@@ -296,23 +381,28 @@ class IntWrapper(ArithmeticTypeWrapper):
                         )
                     )
 
-                return left.convert_to_type(promoteType).convert_bin_op(op, right.convert_to_type(promoteType))
+                return left.convert_to_type(
+                    promoteType, ConversionLevel.Implicit
+                ).convert_bin_op(
+                    op,
+                    right.convert_to_type(promoteType, ConversionLevel.Implicit)
+                )
 
             return super().convert_bin_op(context, left, op, right, inplace)
 
         if op.matches.Mod:
             with context.ifelse(right.nonref_expr) as (ifTrue, ifFalse):
                 with ifFalse:
-                    context.pushException(ZeroDivisionError)
+                    context.pushException(ZeroDivisionError, "division by zero")
 
-            if left.expr_type.typeRepresentation.IsSignedInt:
+            if isSignedInt(left.expr_type.typeRepresentation):
                 return context.pushPod(
                     int,
                     runtime_functions.mod_int64_int64.call(
                         left.toInt64().nonref_expr,
                         right.toInt64().nonref_expr
                     )
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
 
             # unsigned int
             return context.pushPod(
@@ -321,9 +411,9 @@ class IntWrapper(ArithmeticTypeWrapper):
                     left.toUInt64().nonref_expr,
                     right.toUInt64().nonref_expr
                 )
-            ).convert_to_type(self)
+            ).convert_to_type(self, ConversionLevel.Implicit)
         if op.matches.Pow:
-            if left.expr_type.typeRepresentation.IsSignedInt:
+            if isSignedInt(left.expr_type.typeRepresentation):
                 return context.pushPod(
                     float,
                     runtime_functions.pow_int64_int64.call(left.toInt64().nonref_expr, right.toInt64().nonref_expr)
@@ -334,37 +424,37 @@ class IntWrapper(ArithmeticTypeWrapper):
                 runtime_functions.pow_uint64_uint64.call(left.toUInt64().nonref_expr, right.toUInt64().nonref_expr)
             ).toFloat64()
         if op.matches.LShift:
-            if left.expr_type.typeRepresentation.IsSignedInt:
+            if isSignedInt(left.expr_type.typeRepresentation):
                 return context.pushPod(
                     int,
                     runtime_functions.lshift_int64_int64.call(left.toInt64().nonref_expr, right.toInt64().nonref_expr)
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
             # unsigned int
             return context.pushPod(
                 int,
                 runtime_functions.lshift_uint64_uint64.call(left.toUInt64().nonref_expr, right.toUInt64().nonref_expr)
-            ).convert_to_type(self)
+            ).convert_to_type(self, ConversionLevel.Implicit)
         if op.matches.RShift:
-            if left.expr_type.typeRepresentation.IsSignedInt:
+            if isSignedInt(left.expr_type.typeRepresentation):
                 return context.pushPod(
                     int,
                     runtime_functions.rshift_int64_int64.call(left.toInt64().nonref_expr, right.toInt64().nonref_expr)
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
             # unsigned int
             return context.pushPod(
                 int,
                 runtime_functions.rshift_uint64_uint64.call(left.toUInt64().nonref_expr, right.toUInt64().nonref_expr)
-            ).convert_to_type(self)
+            ).convert_to_type(self, ConversionLevel.Implicit)
         if op.matches.FloorDiv:
-            if right.expr_type.typeRepresentation.IsSignedInt:
+            if isSignedInt(right.expr_type.typeRepresentation):
                 return context.pushPod(
                     int,
                     runtime_functions.floordiv_int64_int64.call(left.toInt64().nonref_expr, right.toInt64().nonref_expr)
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
             # unsigned int
             with context.ifelse(right.nonref_expr) as (ifTrue, ifFalse):
                 with ifFalse:
-                    context.pushException(ZeroDivisionError)
+                    context.pushException(ZeroDivisionError, "division by zero")
 
             return context.pushPod(
                 self,
@@ -399,26 +489,30 @@ class IntWrapper(ArithmeticTypeWrapper):
 
 class BoolWrapper(ArithmeticTypeWrapper):
     def __init__(self):
-        super().__init__(Bool)
+        super().__init__(bool)
 
     def getNativeLayoutType(self):
         return native_ast.Type.Int(bits=1, signed=False)
 
     def convert_hash(self, context, expr):
-        return expr.convert_to_type(Int32)
+        return expr.convert_to_type(Int32, ConversionLevel.Implicit)
 
-    def convert_to_type_with_target(self, context, e, targetVal, explicit):
+    def convert_to_type_with_target(self, context, instance, targetVal, conversionLevel, mayThrowOnFailure=False):
         target_type = targetVal.expr_type
 
-        if not explicit:
-            return super().convert_to_type_with_target(context, e, targetVal, explicit)
+        if not isValidConversion(
+            self.typeRepresentation,
+            targetVal.expr_type.typeRepresentation,
+            conversionLevel
+        ):
+            return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
         if isinstance(target_type, FloatWrapper):
             context.pushEffect(
                 targetVal.expr.store(
                     native_ast.Expression.Cast(
-                        left=e.nonref_expr,
-                        to_type=native_ast.Type.Float(bits=target_type.typeRepresentation.Bits)
+                        left=instance.nonref_expr,
+                        to_type=native_ast.Type.Float(bits=bitness(target_type.typeRepresentation))
                     )
                 )
             )
@@ -428,28 +522,27 @@ class BoolWrapper(ArithmeticTypeWrapper):
             context.pushEffect(
                 targetVal.expr.store(
                     native_ast.Expression.Cast(
-                        left=e.nonref_expr,
+                        left=instance.nonref_expr,
                         to_type=native_ast.Type.Int(
-                            bits=target_type.typeRepresentation.Bits,
-                            signed=target_type.typeRepresentation.IsSignedInt
+                            bits=bitness(target_type.typeRepresentation),
+                            signed=signedness(target_type.typeRepresentation)
                         )
                     )
                 )
             )
             return context.constant(True)
 
-        return super().convert_to_type_with_target(context, e, targetVal, explicit)
-
-    def convert_cast(self, context, instance, target_type):
-        if target_type.typeRepresentation == String:
-            return context.push(
-                str,
-                lambda strRef: strRef.expr.store(
-                    runtime_functions.bool_to_string.call(instance.nonref_expr).cast(strRef.expr_type.layoutType)
+        elif target_type.typeRepresentation is str:
+            context.pushEffect(
+                targetVal.expr.store(
+                    runtime_functions.bool_to_string.call(instance.nonref_expr).cast(
+                        targetVal.expr_type.layoutType
+                    )
                 )
             )
+            return context.constant(True)
 
-        return super().convert_cast(context, instance, target_type)
+        return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
     def convert_builtin(self, f, context, expr, a1=None):
         if f is round and a1 is not None:
@@ -468,7 +561,9 @@ class BoolWrapper(ArithmeticTypeWrapper):
 
     def convert_unary_op(self, context, left, op):
         if op.matches.Not:
-            return context.pushPod(self, left.nonref_expr.logical_not())
+            if left.isConstant:
+                return context.constant(not left.constantValue)
+            return context.pushPod(bool, left.nonref_expr.logical_not())
 
         return super().convert_unary_op(context, left, op)
 
@@ -479,7 +574,7 @@ class BoolWrapper(ArithmeticTypeWrapper):
         if op.matches.IsNot and right.expr_type == self:
             op = python_ast.ComparisonOp.NotEq()
 
-        if op.matches.Div:
+        if op.matches.Div and isinstance(right, ArithmeticTypeWrapper):
             T = toWrapper(
                 computeArithmeticBinaryResultType(
                     computeArithmeticBinaryResultType(
@@ -489,7 +584,9 @@ class BoolWrapper(ArithmeticTypeWrapper):
                     Float32
                 )
             )
-            return left.convert_to_type(T).convert_bin_op(op, right.convert_to_type(T))
+            return left.convert_to_type(
+                T, ConversionLevel.Implicit
+            ).convert_bin_op(op, right.convert_to_type(T, ConversionLevel.Implicit))
 
         if right.expr_type != self:
             if isinstance(right.expr_type, ArithmeticTypeWrapper):
@@ -500,7 +597,9 @@ class BoolWrapper(ArithmeticTypeWrapper):
                     )
                 )
 
-                return left.convert_to_type(promoteType).convert_bin_op(op, right.convert_to_type(promoteType))
+                return left.convert_to_type(
+                    promoteType, ConversionLevel.Implicit
+                ).convert_bin_op(op, right.convert_to_type(promoteType, ConversionLevel.Implicit))
 
             return super().convert_bin_op(context, left, op, right, inplace)
 
@@ -515,6 +614,9 @@ class BoolWrapper(ArithmeticTypeWrapper):
                     )
                 )
 
+        if op in pyOpToNative or op.matches.Pow:
+            return left.convert_to_type(int, ConversionLevel.Implicit).convert_bin_op(op, right, inplace)
+
         if op in pyCompOp:
             return context.pushPod(
                 bool,
@@ -533,67 +635,73 @@ class FloatWrapper(ArithmeticTypeWrapper):
         super().__init__(T)
 
     def getNativeLayoutType(self):
-        return native_ast.Type.Float(bits=self.typeRepresentation.Bits)
+        return native_ast.Type.Float(bits=bitness(self.typeRepresentation))
 
     def convert_hash(self, context, expr):
         if self.typeRepresentation == Float32:
             return context.pushPod(Int32, runtime_functions.hash_float32.call(expr.nonref_expr))
-        if self.typeRepresentation == Float64:
+        if self.typeRepresentation == float:
             return context.pushPod(Int32, runtime_functions.hash_float64.call(expr.nonref_expr))
 
         assert False
 
-    def convert_to_type_with_target(self, context, e, targetVal, explicit):
+    def convert_to_type_with_target(self, context, instance, targetVal, conversionLevel, mayThrowOnFailure=False):
         target_type = targetVal.expr_type
 
-        if not explicit:
-            return super().convert_to_type_with_target(context, e, targetVal, explicit)
+        if not isValidConversion(
+            self.typeRepresentation,
+            targetVal.expr_type.typeRepresentation,
+            conversionLevel
+        ):
+            return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
         if isinstance(target_type, FloatWrapper):
             context.pushEffect(
                 targetVal.expr.store(
                     native_ast.Expression.Cast(
-                        left=e.nonref_expr,
-                        to_type=native_ast.Type.Float(bits=target_type.typeRepresentation.Bits)
+                        left=instance.nonref_expr,
+                        to_type=native_ast.Type.Float(bits=bitness(target_type.typeRepresentation))
                     )
                 )
             )
             return context.constant(True)
 
-        if target_type.typeRepresentation == Bool:
-            targetVal.convert_copy_initialize(e != 0.0)
+        if isinstance(target_type, BoolWrapper):
+            context.pushEffect(
+                targetVal.expr.store(
+                    instance.nonref_expr.neq(instance.expr_type.getNativeLayoutType().zero())
+                )
+            )
             return context.constant(True)
 
         if isinstance(target_type, IntWrapper):
-            context.pushEffect(
-                targetVal.expr.store(
-                    native_ast.Expression.Cast(
-                        left=e.nonref_expr,
-                        to_type=native_ast.Type.Int(
-                            bits=target_type.typeRepresentation.Bits,
-                            signed=target_type.typeRepresentation.IsSignedInt
-                        )
-                    )
+            if self.typeRepresentation == float:
+                func = runtime_functions.float64_to_int
+            else:
+                func = runtime_functions.float32_to_int
+
+            return context.pushPod(
+                bool,
+                func.call(
+                    targetVal.expr.cast(native_ast.VoidPtr),
+                    instance.nonref_expr,
+                    native_ast.const_bool_expr(mayThrowOnFailure),
+                    context.getTypePointer(targetVal.expr_type.typeRepresentation)
                 )
             )
-            return context.constant(True)
 
-        return super().convert_to_type_with_target(context, e, targetVal, explicit)
-
-    def convert_cast(self, context, instance, target_type):
-        if target_type.typeRepresentation == String:
-            if self.typeRepresentation == Float64:
+        if target_type.typeRepresentation is str:
+            if self.typeRepresentation == float:
                 func = runtime_functions.float64_to_string
             else:
                 func = runtime_functions.float32_to_string
 
-            return context.push(
-                str,
-                lambda targetVal:
-                    targetVal.expr.store(func.call(instance.nonref_expr).cast(target_type.layoutType))
+            context.pushEffect(
+                targetVal.expr.store(func.call(instance.nonref_expr).cast(targetVal.expr_type.layoutType))
             )
+            return context.constant(True)
 
-        return super().convert_cast(context, instance, target_type)
+        return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
     def convert_abs(self, context, expr):
         return context.pushPod(
@@ -611,27 +719,40 @@ class FloatWrapper(ArithmeticTypeWrapper):
                 return context.pushPod(
                     float,
                     runtime_functions.round_float64.call(expr.toFloat64().nonref_expr, a1.toInt64().nonref_expr)
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
             else:
                 return context.pushPod(
                     float,
                     runtime_functions.round_float64.call(expr.toFloat64().nonref_expr, context.constant(0))
-                ).convert_to_type(self)
+                ).convert_to_type(self, ConversionLevel.Implicit)
         if f is trunc:
-            return context.pushPod(float, runtime_functions.trunc_float64.call(expr.toFloat64().nonref_expr)).convert_to_type(self)
+            return context.pushPod(
+                float,
+                runtime_functions.trunc_float64.call(expr.toFloat64().nonref_expr)
+            )
         if f is floor:
-            return context.pushPod(float, runtime_functions.floor_float64.call(expr.toFloat64().nonref_expr)).convert_to_type(self)
+            return context.pushPod(
+                float, runtime_functions.floor_float64.call(expr.toFloat64().nonref_expr)
+            )
         if f is ceil:
-            return context.pushPod(float, runtime_functions.ceil_float64.call(expr.toFloat64().nonref_expr)).convert_to_type(self)
+            return context.pushPod(
+                float, runtime_functions.ceil_float64.call(expr.toFloat64().nonref_expr)
+            )
 
         return super().convert_builtin(f, context, expr, a1)
 
     def convert_unary_op(self, context, left, op):
         if op.matches.Not:
-            return context.pushPod(self, left.nonref_expr.logical_not())
+            if left.isConstant:
+                return context.constant(not left.constantValue)
+            return context.pushPod(bool, left.nonref_expr.eq(self.getNativeLayoutType().zero()))
         if op.matches.USub:
+            if left.isConstant:
+                return context.constant(-left.constantValue)
             return context.pushPod(self, left.nonref_expr.negate())
         if op.matches.UAdd:
+            if left.isConstant:
+                return context.constant(left.constantValue)
             return context.pushPod(self, left.nonref_expr)
 
         return super().convert_unary_op(context, left, op)
@@ -640,7 +761,7 @@ class FloatWrapper(ArithmeticTypeWrapper):
         if right.expr_type != self:
             if isinstance(right.expr_type, ArithmeticTypeWrapper):
                 if op.matches.Pow:
-                    promoteType = toWrapper(Float64)
+                    promoteType = toWrapper(float)
                 else:
                     promoteType = toWrapper(
                         computeArithmeticBinaryResultType(
@@ -648,18 +769,20 @@ class FloatWrapper(ArithmeticTypeWrapper):
                             right.expr_type.typeRepresentation
                         )
                     )
-                return left.convert_to_type(promoteType).convert_bin_op(op, right.convert_to_type(promoteType))
+                return left.convert_to_type(
+                    promoteType, ConversionLevel.Implicit
+                ).convert_bin_op(op, right.convert_to_type(promoteType, ConversionLevel.Implicit))
             return super().convert_bin_op(context, left, op, right, inplace)
 
         if op.matches.Mod:
             # TODO: might define mod_float32_float32 instead of doing these conversions
             if left.expr_type.typeRepresentation == Float32:
                 return left.toFloat64().convert_bin_op(
-                    op, right.toFloat64()).convert_to_type(toWrapper(Float32))
+                    op, right.toFloat64()).convert_to_type(Float32, ConversionLevel.Implicit)
 
             with context.ifelse(right.nonref_expr) as (ifTrue, ifFalse):
                 with ifFalse:
-                    context.pushException(ZeroDivisionError)
+                    context.pushException(ZeroDivisionError, "division by zero")
 
             return context.pushPod(
                 self,
@@ -667,9 +790,15 @@ class FloatWrapper(ArithmeticTypeWrapper):
             )
 
         if op.matches.Div:
+            if left.isConstant and right.isConstant:
+                # If we divide by 0, for compatibility, we want that to remain a runtime error, not a compile error.
+                try:
+                    return context.constant(left.constantValue / right.constantValue).convert_to_type(self, ConversionLevel.Implicit)
+                except Exception:
+                    pass
             with context.ifelse(right.nonref_expr) as (ifTrue, ifFalse):
                 with ifFalse:
-                    context.pushException(ZeroDivisionError)
+                    context.pushException(ZeroDivisionError, "division by zero")
 
             return context.pushPod(
                 self,
@@ -689,7 +818,7 @@ class FloatWrapper(ArithmeticTypeWrapper):
             return context.pushPod(
                 float,
                 runtime_functions.floordiv_float64_float64.call(left.toFloat64().nonref_expr, right.toFloat64().nonref_expr)
-            ).convert_to_type(self)
+            ).convert_to_type(self, ConversionLevel.Implicit)
 
         if op in pyOpToNative and op not in pyOpNotForFloat:
             return context.pushPod(
