@@ -1,4 +1,4 @@
-#   Copyright 2017-2019 typed_python Authors
+#   Copyright 2017-2020 typed_python Authors
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
@@ -15,11 +15,12 @@
 from typed_python.compiler.type_wrappers.wrapper import Wrapper
 from typed_python.compiler.type_wrappers.refcounted_wrapper import RefcountedWrapper
 import typed_python.compiler.type_wrappers.runtime_functions as runtime_functions
-from typed_python.compiler.type_wrappers.bound_compiled_method_wrapper import BoundCompiledMethodWrapper
+from typed_python.compiler.conversion_level import ConversionLevel
+from typed_python.compiler.type_wrappers.bound_method_wrapper import BoundMethodWrapper
 from typed_python.compiler.type_wrappers.util import min
 from typed_python.compiler.typed_expression import TypedExpression
 
-from typed_python import NoneType, Tuple, Bool
+from typed_python import Tuple
 
 import typed_python.compiler.native_ast as native_ast
 import typed_python.compiler
@@ -106,7 +107,27 @@ def const_dict_getitem(constDict, key):
         else:
             return constDict.get_value_by_index_unsafe(mid)
 
-    raise Exception("Key doesn't exist")
+    raise KeyError(key)
+
+
+def const_dict_get(constDict, key, default):
+    # perform a binary search
+    lowIx = 0
+    highIx = len(constDict)
+
+    while lowIx < highIx:
+        mid = (lowIx + highIx) >> 1
+
+        keyAtVal = constDict.get_key_by_index_unsafe(mid)
+
+        if keyAtVal < key:
+            lowIx = mid + 1
+        elif key < keyAtVal:
+            highIx = mid
+        else:
+            return constDict.get_value_by_index_unsafe(mid)
+
+    return default
 
 
 def const_dict_contains(constDict, key):
@@ -129,10 +150,6 @@ def const_dict_contains(constDict, key):
     return False
 
 
-def const_dict_contains_not(constDict, key):
-    return False if const_dict_contains(constDict, key) else True
-
-
 class ConstDictWrapperBase(RefcountedWrapper):
     """Common method wrappers for all ConstDicts.
 
@@ -177,7 +194,7 @@ class ConstDictWrapperBase(RefcountedWrapper):
                     "destructor_" + str(self.constDictType),
                     ('destructor', self),
                     [self],
-                    typeWrapper(NoneType),
+                    typeWrapper(type(None)),
                     self.generateNativeDestructorFunction
                 )
                 .call(instance)
@@ -198,8 +215,8 @@ class ConstDictWrapper(ConstDictWrapperBase):
         super().__init__(constDictType, None)
 
     def convert_attribute(self, context, instance, attr):
-        if attr in ("get_key_by_index_unsafe", "get_value_by_index_unsafe", "keys", "values", "items"):
-            return instance.changeType(BoundCompiledMethodWrapper(self, attr))
+        if attr in ("get_key_by_index_unsafe", "get_value_by_index_unsafe", "keys", "values", "items", "get"):
+            return instance.changeType(BoundMethodWrapper.Make(self, attr))
 
         return super().convert_attribute(context, instance, attr)
 
@@ -220,6 +237,12 @@ class ConstDictWrapper(ConstDictWrapperBase):
 
             return res
 
+        if methodname == "get" and not kwargs:
+            if len(args) == 1:
+                return self.convert_get(context, instance, args[0], context.constant(None))
+            elif len(args) == 2:
+                return self.convert_get(context, instance, args[0], args[1])
+
         if methodname == "keys" and not args and not kwargs:
             return instance.changeType(ConstDictKeysWrapper(self.constDictType))
 
@@ -293,26 +316,32 @@ class ConstDictWrapper(ConstDictWrapperBase):
         return super().convert_bin_op(context, left, op, right, inplace)
 
     def convert_bin_op_reverse(self, context, left, op, right, inplace):
-        if op.matches.In or op.matches.NotIn:
-            right = right.convert_to_type(self.keyType)
+        if op.matches.In:
+            right = right.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
             if right is None:
                 return None
 
-            return context.call_py_function(
-                const_dict_contains if op.matches.In else const_dict_contains_not,
-                (left, right),
-                {}
-            )
+            return context.call_py_function(const_dict_contains, (left, right), {})
 
-        return super().convert_bin_op(context, left, op, right, inplace)
+        return super().convert_bin_op_reverse(context, left, op, right, inplace)
 
     def convert_getitem(self, context, instance, item):
-        item = item.convert_to_type(self.keyType)
+        item = item.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
         if item is None:
             return None
 
         return context.call_py_function(const_dict_getitem, (instance, item), {})
 
+    def convert_get(self, context, expr, item, default):
+        if item is None or expr is None or default is None:
+            return None
+
+        item = item.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
+        if item is None:
+            return None
+
+        return context.call_py_function(const_dict_get, (expr, item, default), {})
+
     def convert_len_native(self, expr):
         if isinstance(expr, TypedExpression):
             expr = expr.nonref_expr
@@ -325,21 +354,27 @@ class ConstDictWrapper(ConstDictWrapperBase):
     def convert_len(self, context, expr):
         return context.pushPod(int, self.convert_len_native(expr.nonref_expr))
 
-    def convert_to_type_with_target(self, context, e, targetVal, explicit):
-        if not explicit:
-            return super().convert_to_type_with_target(context, e, targetVal, explicit)
+    def _can_convert_to_type(self, targetType, conversionLevel):
+        if not conversionLevel.isNewOrHigher():
+            return False
 
-        target_type = targetVal.expr_type
+        if targetType.typeRepresentation is bool:
+            return True
 
-        if target_type.typeRepresentation == Bool:
+        if targetType.typeRepresentation is str:
+            return "Maybe"
+
+        return False
+
+    def convert_to_type_with_target(self, context, instance, targetVal, conversionLevel, mayThrowOnFailure=False):
+        if targetVal.expr_type.typeRepresentation is bool:
+            res = context.pushPod(bool, self.convert_len_native(instance.nonref_expr).neq(0))
             context.pushEffect(
-                targetVal.expr.store(
-                    self.convert_len_native(e).neq(0)
-                )
+                targetVal.expr.store(res.nonref_expr)
             )
             return context.constant(True)
 
-        return super().convert_to_type_with_target(context, e, targetVal, explicit)
+        return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
 
 
 class ConstDictMakeIteratorWrapper(ConstDictWrapperBase):
@@ -397,7 +432,7 @@ class ConstDictIteratorWrapper(Wrapper):
             name="const_dict_iterator"
         )
 
-    def convert_next(self, context, expr):
+    def convert_fastnext(self, context, expr):
         context.pushEffect(
             expr.expr.ElementPtrIntegers(0, 0).store(
                 expr.expr.ElementPtrIntegers(0, 0).load().add(1)
@@ -411,7 +446,7 @@ class ConstDictIteratorWrapper(Wrapper):
 
         nextIx = context.pushReference(int, expr.expr.ElementPtrIntegers(0, 0))
 
-        return self.iteratedItemForReference(context, expr, nextIx), canContinue
+        return self.iteratedItemForReference(context, expr, nextIx).asPointerIf(canContinue)
 
     def refAs(self, context, expr, which):
         assert expr.expr_type == self
