@@ -1,4 +1,4 @@
-#   Copyright 2018 Braxton Mckee
+#   Copyright 2017-2020 typed_python Authors
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
@@ -14,11 +14,13 @@
 
 from typed_python.compiler.type_wrappers.refcounted_wrapper import RefcountedWrapper
 from typed_python.compiler.typed_expression import TypedExpression
+from typed_python.compiler.conversion_level import ConversionLevel
 import typed_python.compiler.type_wrappers.runtime_functions as runtime_functions
-from typed_python.compiler.type_wrappers.bound_compiled_method_wrapper import BoundCompiledMethodWrapper
+from typed_python.compiler.type_wrappers.bound_method_wrapper import BoundMethodWrapper
 from typed_python.compiler.type_wrappers.wrapper import Wrapper
-from typed_python.compiler.type_wrappers.compilable_builtin import CompilableBuiltin
-from typed_python import NoneType, Tuple, PointerTo, Int32, Int64, UInt8, Bool
+from typed_python.compiler.type_wrappers.native_hash import table_next_slot, table_clear, table_contains, \
+    dict_delitem, dict_getitem, dict_get, dict_setitem
+from typed_python import Tuple, PointerTo, Int32, UInt8, Dict, ConstDict
 
 import typed_python.compiler.native_ast as native_ast
 import typed_python.compiler
@@ -26,192 +28,86 @@ import typed_python.compiler
 
 typeWrapper = lambda t: typed_python.compiler.python_object_representation.typedPythonTypeToTypeWrapper(t)
 
-EMPTY = -1
-DELETED = -2
+
+def checkDictSizeAndThrowIfChanged(instance, expectedLen):
+    if len(instance) != expectedLen:
+        raise RuntimeError("dictionary size changed during iteration")
 
 
-class CPlusPlusStyleMod(CompilableBuiltin):
-    def convert_call(self, context, instance, args, kwargs):
-        if len(args) == 2 and args[0].expr_type.typeRepresentation == Int32 and args[1].expr_type.typeRepresentation == Int64:
-            return context.pushPod(
-                int,
-                args[0].nonref_expr.cast(native_ast.Int64).mod(args[1].nonref_expr)
-            )
-
-        if len(args) == 2 and args[0].expr_type.typeRepresentation == Int64 and args[1].expr_type.typeRepresentation == Int64:
-            return context.pushPod(
-                int,
-                args[0].nonref_expr.mod(args[1].nonref_expr)
-            )
-
-        return super().convert_call(context, instance, args, kwargs)
+def dict_update(instance, other):
+    for key in other:
+        instance[key] = other[key]
 
 
-def dict_add_slot(instance, itemHash, slot):
-    if (instance._hash_table_count * 2 + 1 > instance._hash_table_size or
-            instance._hash_table_empty_slots < instance._hash_table_size >> 2 + 1):
-        instance._resizeTableUnsafe()
+def dict_setdefault_nodefault(dict, item):
+    if item not in dict:
+        dict[item] = type(dict).ValueType()
+    return dict[item]
 
-    if itemHash < 0:
-        itemHash = -itemHash
 
-    offset = CPlusPlusStyleMod(itemHash, instance._hash_table_size)
-
-    while True:
-        if instance._hash_table_slots[offset] == EMPTY or instance._hash_table_slots[offset] == DELETED:
-            if instance._hash_table_slots[offset] == EMPTY:
-                instance._hash_table_empty_slots -= 1
-
-            instance._hash_table_slots[offset] = slot
-            instance._hash_table_hashes[offset] = itemHash
-            instance._items_populated[slot] = 1
-            instance._hash_table_count += 1
-
-            return
-
-        offset += 1
-
-        if offset >= instance._hash_table_size:
-            offset = 0
-
-
-def dict_slot_for_key(instance, itemHash, item):
-    slots = instance._hash_table_slots
-
-    if not slots:
-        return -1
-
-    if itemHash < 0:
-        itemHash = -itemHash
-
-    offset = CPlusPlusStyleMod(itemHash, instance._hash_table_size)
-
-    while True:
-        slotIndex = int((slots + offset).get())
-
-        if slotIndex == EMPTY:
-            return -1
-
-        if slotIndex != DELETED and (instance._hash_table_hashes + offset).get() == itemHash:
-            if instance.getKeyByIndexUnsafe(slotIndex) == item:
-                return slotIndex
-
-        offset += 1
-        if offset >= instance._hash_table_size:
-            offset = 0
-
-    # not necessary, but currently we don't realize that the while loop
-    # never exits, and so we think there's a possibility we return None
-    return 0
-
-
-def dict_next_slot(instance, slotIx):
-    slotIx += 1
-
-    while slotIx < instance._items_reserved:
-        if instance._items_populated[slotIx]:
-            return slotIx
-        slotIx += 1
-
-    return -1
-
-
-def dict_remove_key(instance, item, itemHash):
-    slots = instance._hash_table_slots
-
-    if not slots:
-        raise KeyError(item)
-
-    if instance._items_reserved > (instance._hash_table_count + 2) * 4:
-        instance._compressItemTableUnsafe()
-
-    if instance._hash_table_count < instance._hash_table_size >> 3:
-        instance._resizeTableUnsafe()
-
-    if itemHash < 0:
-        itemHash = -itemHash
-
-    offset = CPlusPlusStyleMod(itemHash, instance._hash_table_size)
-
-    while True:
-        slotIndex = int((slots + offset).get())
-
-        if slotIndex == EMPTY:
-            raise KeyError(item)
-
-        if slotIndex != DELETED and (instance._hash_table_hashes + offset).get() == itemHash:
-            if instance.getKeyByIndexUnsafe(slotIndex) == item:
-                instance._hash_table_hashes[offset] = -1
-                instance._hash_table_slots[offset] = DELETED
-                instance._hash_table_count -= 1
-                instance._items_populated[slotIndex] = 0
-
-                instance.deleteItemByIndexUnsafe(slotIndex)
-                return
-
-        offset += 1
-        if offset >= instance._hash_table_size:
-            offset = 0
-
-    # not necessary, but currently we don't currently realize that the while loop
-    # never exits, and so we think there's a possibility we return None
-    return 0
-
-
-def dict_delitem(instance, item):
-    itemHash = hash(item)
-
-    dict_remove_key(instance, item, itemHash)
-
-
-def dict_getitem(instance, item):
-    itemHash = hash(item)
-
-    slot = dict_slot_for_key(instance, itemHash, item)
-
-    if slot == -1:
-        raise KeyError(item)
-
-    return instance.getValueByIndexUnsafe(slot)
-
-
-def dict_contains(instance, item):
-    itemHash = hash(item)
-
-    slot = dict_slot_for_key(instance, itemHash, item)
-
-    return slot != -1
-
-
-def dict_contains_not(instance, item):
-    itemHash = hash(item)
-
-    slot = dict_slot_for_key(instance, itemHash, item)
-
-    return slot == -1
-
-
-def dict_setitem(instance, key, value):
-    itemHash = hash(key)
-
-    slot = dict_slot_for_key(instance, itemHash, key)
-
-    if slot == -1:
-        newSlot = instance._allocateNewSlotUnsafe()
-        dict_add_slot(instance, itemHash, newSlot)
-        instance.initializeKeyByIndexUnsafe(newSlot, key)
-        instance.initializeValueByIndexUnsafe(newSlot, value)
-    else:
-        instance.assignValueByIndexUnsafe(slot, value)
-
-
-def dict_setdefault(dict, item, defaultValue=None):
+def dict_setdefault(dict, item, defaultValue):
     if item not in dict:
-        # TypeError: Can't convert from type String to type Int64
         dict[item] = defaultValue
     return dict[item]
 
 
+def initialize_dict_from_mappable(ptrToOutDict, mappable, mayThrow):
+    """Initialize a dictionary from an arbitrary object supporting the mapping protocol.
+
+    This is called internally by the compiler to support initializing
+    a Dict(T1, T2) from another Dict with different key/value types.
+
+    Args:
+        ptrToOutDict - a pointer to an uninitialized Dict instance.
+        mappable - something we can iterate over with a getitem.
+    Returns:
+        True if we succeeded, and ptrToOutDict is valid. False if
+        we failed, and ptrToOutDict will not point to an initialized
+        dictionary.
+    """
+    ptrToOutDict.initialize()
+
+    try:
+        for m in mappable:
+            ptrToOutDict.get()[m] = mappable[m]
+
+        return True
+    except: # noqa
+        ptrToOutDict.destroy()
+        if mayThrow:
+            raise
+        return False
+
+
+def dict_pop_nodefault(dict, item):
+    if item not in dict:
+        raise KeyError(item)
+
+    res = dict[item]
+
+    del dict[item]
+
+    return res
+
+
+def dict_pop(dict, item, defaultValue):
+    if item not in dict:
+        return defaultValue
+
+    res = dict[item]
+    del dict[item]
+    return res
+
+
+def dict_duplicate(d):
+    out = type(d)()
+
+    for k in d.items():
+        out[k[0]] = k[1]
+
+    return out
+
+
 class DictWrapperBase(RefcountedWrapper):
     is_pod = False
     is_empty = False
@@ -253,7 +149,7 @@ class DictWrapperBase(RefcountedWrapper):
                 "destructor_" + str(self.typeRepresentation),
                 ('destructor', self),
                 [self],
-                typeWrapper(NoneType),
+                typeWrapper(type(None)),
                 self.generateNativeDestructorFunction
             ).call(instance)
         )
@@ -261,6 +157,11 @@ class DictWrapperBase(RefcountedWrapper):
     def getNativeLayoutType(self):
         return self.layoutType
 
+    def convert_len_native(self, expr):
+        if isinstance(expr, TypedExpression):
+            expr = expr.nonref_expr
+        return expr.ElementPtrIntegers(0, 8).load().cast(native_ast.Int64)
+
 
 class DictWrapper(DictWrapperBase):
     def __init__(self, dictType):
@@ -269,7 +170,7 @@ class DictWrapper(DictWrapperBase):
     def convert_default_initialize(self, context, instance):
         context.pushEffect(
             instance.expr.store(
-                runtime_functions.dict_create.call().cast(self.layoutType)
+                runtime_functions.table_create.call().cast(self.layoutType)
             )
         )
 
@@ -278,8 +179,9 @@ class DictWrapper(DictWrapperBase):
                 "getItemByIndexUnsafe", "getKeyByIndexUnsafe", "getValueByIndexUnsafe", "deleteItemByIndexUnsafe",
                 "initializeValueByIndexUnsafe", "assignValueByIndexUnsafe",
                 "initializeKeyByIndexUnsafe", "_allocateNewSlotUnsafe", "_resizeTableUnsafe",
-                "_compressItemTableUnsafe", "get", "items", "keys", "values", "setdefault"):
-            return expr.changeType(BoundCompiledMethodWrapper(self, attr))
+                "_top_item_slot", "_compressItemTableUnsafe", "get", "items", "keys", "values", "setdefault",
+                "pop", "clear", "copy", "update"):
+            return expr.changeType(BoundMethodWrapper.Make(self, attr))
 
         if attr == '_items_populated':
             return context.pushPod(
@@ -289,10 +191,16 @@ class DictWrapper(DictWrapperBase):
 
         if attr == '_items_reserved':
             return context.pushPod(
-                Int64,
+                int,
                 expr.nonref_expr.ElementPtrIntegers(0, 3).load()
             )
 
+        if attr == '_top_item_slot':
+            return context.pushPod(
+                int,
+                expr.nonref_expr.ElementPtrIntegers(0, 4).load()
+            )
+
         if attr == '_hash_table_slots':
             return context.pushPod(
                 PointerTo(Int32),
@@ -325,9 +233,42 @@ class DictWrapper(DictWrapperBase):
 
         return super().convert_attribute(context, expr, attr)
 
+    @staticmethod
+    def supportsMappingProtocol(T):
+        """Is T compiled and supporting the mapping protocol?"""
+        if issubclass(T, (Dict, ConstDict)):
+            return True
+
+        # when we support iterators in Class instances, we can revisit this:
+        # if issubclass(T, Class) and '__getitem__' in T.Methods and '__iter__' in T.Methods
+        #    return False
+
+        return False
+
+    def _can_convert_from_type(self, otherType, conversionLevel):
+        if (
+            conversionLevel.isImplicitContainersOrHigher()
+            and DictWrapper.supportsMappingProtocol(otherType.typeRepresentation)
+        ):
+            return "Maybe"
+
+        return super()._can_convert_from_type(otherType, conversionLevel)
+
+    def convert_to_self_with_target(self, context, targetVal, sourceVal, conversionLevel, mayThrowOnFailure=False):
+        if conversionLevel.isImplicitContainersOrHigher() and DictWrapper.supportsMappingProtocol(sourceVal.expr_type.typeRepresentation):
+            res = context.call_py_function(
+                initialize_dict_from_mappable,
+                (targetVal.asPointer(), sourceVal, context.constant(mayThrowOnFailure)),
+                {}
+            )
+
+            return res
+
+        return super().convert_to_self_with_target(context, targetVal, sourceVal, conversionLevel, mayThrowOnFailure)
+
     def convert_set_attribute(self, context, instance, attr, expr):
         if attr == '_hash_table_count':
-            val = expr.convert_to_type(int)
+            val = expr.toInt64()
             if val is None:
                 return None
             context.pushEffect(
@@ -336,8 +277,19 @@ class DictWrapper(DictWrapperBase):
 
             return context.pushVoid()
 
+        if attr == '_top_item_slot':
+            val = expr.toInt64()
+            if val is None:
+                return None
+
+            context.pushEffect(
+                instance.nonref_expr.ElementPtrIntegers(0, 4).store(val.nonref_expr)
+            )
+
+            return context.pushVoid()
+
         if attr == '_hash_table_empty_slots':
-            val = expr.convert_to_type(int)
+            val = expr.toInt64()
             if val is None:
                 return None
             context.pushEffect(
@@ -355,15 +307,18 @@ class DictWrapper(DictWrapperBase):
         if methodname == "__iter__" and not args and not kwargs:
             res = context.push(
                 DictKeysIteratorWrapper(self.dictType),
-                lambda instance:
-                    instance.expr.ElementPtrIntegers(0, 0).store(-1)
-                    # we initialize the dict pointer below, so technically
-                    # if that were to throw, this would leak a bad value.
+                lambda iteratorInstance:
+                iteratorInstance.expr.ElementPtrIntegers(0, 0).store(-1)
+                >> iteratorInstance.expr.ElementPtrIntegers(0, 1).store(
+                    self.convert_len_native(instance)
+                )
+                # we initialize the dict pointer below, so technically
+                # if that were to throw, this would leak a bad value.
             )
 
             context.pushReference(
                 self,
-                res.expr.ElementPtrIntegers(0, 1)
+                res.expr.ElementPtrIntegers(0, 2)
             ).convert_copy_initialize(instance)
 
             return res
@@ -378,9 +333,19 @@ class DictWrapper(DictWrapperBase):
             return instance.changeType(DictItemsWrapper(self.dictType))
 
         if len(args) == 0:
+            if methodname == "copy":
+                return context.push(
+                    typeWrapper(self.dictType),
+                    lambda ref: ref.expr.store(
+                        runtime_functions.table_copy.call(
+                            instance.nonref_expr.cast(native_ast.VoidPtr),
+                            context.getTypePointer(self.dictType)
+                        ).cast(self.layoutType)
+                    )
+                )
             if methodname == "_compressItemTableUnsafe":
                 context.pushEffect(
-                    runtime_functions.dict_compressItemTable.call(
+                    runtime_functions.table_compress.call(
                         instance.nonref_expr.cast(native_ast.VoidPtr),
                         context.constant(self.kvBytecount)
                     )
@@ -389,7 +354,7 @@ class DictWrapper(DictWrapperBase):
 
             if methodname == "_resizeTableUnsafe":
                 context.pushEffect(
-                    runtime_functions.dict_resizeTable.call(
+                    runtime_functions.table_resize.call(
                         instance.nonref_expr.cast(native_ast.VoidPtr)
                     )
                 )
@@ -398,7 +363,7 @@ class DictWrapper(DictWrapperBase):
             if methodname == "_allocateNewSlotUnsafe":
                 return context.pushPod(
                     Int32,
-                    runtime_functions.dict_allocateNewSlot.call(
+                    runtime_functions.table_allocate_new_slot.call(
                         instance.nonref_expr.cast(native_ast.VoidPtr),
                         context.constant(self.kvBytecount)
                     )
@@ -406,16 +371,30 @@ class DictWrapper(DictWrapperBase):
 
         if methodname == "setdefault":
             if len(args) == 1:
-                return context.call_py_function(dict_setdefault, (instance, args[0]), {})
+                return context.call_py_function(dict_setdefault_nodefault, (instance, args[0]), {})
             else:
                 return context.call_py_function(dict_setdefault, (instance, args[0], args[1]), {})
 
+        if methodname == "pop":
+            if len(args) == 1:
+                return context.call_py_function(dict_pop_nodefault, (instance, args[0]), {})
+            else:
+                return context.call_py_function(dict_pop, (instance, args[0], args[1]), {})
+
+        if methodname == "clear":
+            if len(args) == 0:
+                return context.call_py_function(table_clear, (instance,), {})
+
+        if methodname == "update":
+            if len(args) == 1:
+                return context.call_py_function(dict_update, (instance, args[0]), {})
+
         if len(args) == 1:
             if methodname == "get":
-                return self.convert_getitem(context, instance, args[0])
+                return self.convert_get(context, instance, args[0], context.constant(None))
 
             if methodname in ("getItemByIndexUnsafe", "getKeyByIndexUnsafe", "getValueByIndexUnsafe", "deleteItemByIndexUnsafe"):
-                index = args[0].convert_to_type(int)
+                index = args[0].toInt64()
                 if index is None:
                     return None
 
@@ -439,12 +418,14 @@ class DictWrapper(DictWrapperBase):
                     return item.expr_type.refAs(context, item, 1)
 
         if len(args) == 2:
+            if methodname == "get":
+                return self.convert_get(context, instance, args[0], args[1])
             if methodname in ("initializeValueByIndexUnsafe", 'assignValueByIndexUnsafe'):
-                index = args[0].convert_to_type(int)
+                index = args[0].toInt64()
                 if index is None:
                     return None
 
-                value = args[1].convert_to_type(self.valueType)
+                value = args[1].convert_to_type(self.valueType, ConversionLevel.Implicit)
                 if value is None:
                     return None
 
@@ -463,11 +444,11 @@ class DictWrapper(DictWrapperBase):
                 return context.pushVoid()
 
             if methodname == "initializeKeyByIndexUnsafe":
-                index = args[0].convert_to_type(int)
+                index = args[0].toInt64()
                 if index is None:
                     return None
 
-                key = args[1].convert_to_type(self.keyType)
+                key = args[1].convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
                 if key is None:
                     return None
 
@@ -488,7 +469,7 @@ class DictWrapper(DictWrapperBase):
         if item is None or expr is None:
             return None
 
-        item = item.convert_to_type(self.keyType)
+        item = item.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
         if item is None:
             return None
 
@@ -498,31 +479,36 @@ class DictWrapper(DictWrapperBase):
         if item is None or expr is None:
             return None
 
-        item = item.convert_to_type(self.keyType)
+        item = item.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
         if item is None:
             return None
 
         return context.call_py_function(dict_getitem, (expr, item), {})
 
+    def convert_get(self, context, expr, item, default):
+        if item is None or expr is None:
+            return None
+
+        item = item.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
+        if item is None:
+            return None
+
+        return context.call_py_function(dict_get, (expr, item, default), {})
+
     def convert_setitem(self, context, expr, key, value):
         if key is None or expr is None or value is None:
             return None
 
-        key = key.convert_to_type(self.keyType)
+        key = key.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
         if key is None:
             return None
 
-        value = value.convert_to_type(self.valueType)
+        value = value.convert_to_type(self.valueType, ConversionLevel.ImplicitContainers)
         if value is None:
             return None
 
         return context.call_py_function(dict_setitem, (expr, key, value), {})
 
-    def convert_len_native(self, expr):
-        if isinstance(expr, TypedExpression):
-            expr = expr.nonref_expr
-        return expr.ElementPtrIntegers(0, 8).load().cast(native_ast.Int64)
-
     def convert_items_reserved_native(self, expr):
         if isinstance(expr, TypedExpression):
             expr = expr.nonref_expr
@@ -540,18 +526,14 @@ class DictWrapper(DictWrapperBase):
         return context.pushPod(int, self.convert_len_native(expr))
 
     def convert_bin_op_reverse(self, context, left, op, right, inplace):
-        if op.matches.In or op.matches.NotIn:
-            right = right.convert_to_type(self.keyType)
+        if op.matches.In:
+            right = right.convert_to_type(self.keyType, ConversionLevel.UpcastContainers)
             if right is None:
                 return None
 
-            return context.call_py_function(
-                dict_contains if op.matches.In else dict_contains_not,
-                (left, right),
-                {}
-            )
+            return context.call_py_function(table_contains, (left, right), {})
 
-        return super().convert_bin_op(context, left, op, right, inplace)
+        return super().convert_bin_op_reverse(context, left, op, right, inplace)
 
     def convert_getkey_by_index_unsafe(self, context, expr, item):
         return context.pushReference(
@@ -586,46 +568,58 @@ class DictWrapper(DictWrapperBase):
             runtime_functions.free.call(inst.nonref_expr.cast(native_ast.UInt8Ptr))
         )
 
-    def convert_to_type_with_target(self, context, e, targetVal, explicit):
-        if not explicit:
-            return super().convert_to_type_with_target(context, e, targetVal, explicit)
-
-        target_type = targetVal.expr_type
-
-        if target_type.typeRepresentation == Bool:
-            context.pushEffect(
-                targetVal.expr.store(
-                    self.convert_len_native(e.nonref_expr).neq(0)
-                )
-            )
-            return context.constant(True)
-
-        return super().convert_to_type_with_target(context, e, targetVal, explicit)
-
     def convert_type_call(self, context, typeInst, args, kwargs):
         if len(args) == 0 and not kwargs:
             return context.push(self, lambda x: x.convert_default_initialize())
 
         if len(args) == 1 and not kwargs:
-            return args[0].convert_to_type(self, True)
+            if args[0].expr_type == self:
+                return context.call_py_function(dict_duplicate, (args[0],), {})
+            else:
+                return args[0].convert_to_type(self, ConversionLevel.New)
 
         return super().convert_type_call(context, typeInst, args, kwargs)
 
+    def _can_convert_to_type(self, targetType, conversionLevel):
+        if not conversionLevel.isNewOrHigher():
+            return False
+
+        if targetType.typeRepresentation is bool:
+            return True
+
+        if targetType.typeRepresentation is str:
+            return "Maybe"
+
+        return False
+
+    def convert_to_type_with_target(self, context, instance, targetVal, conversionLevel, mayThrowOnFailure=False):
+        if targetVal.expr_type.typeRepresentation is bool:
+            res = context.pushPod(bool, self.convert_len_native(instance.nonref_expr).neq(0))
+            context.pushEffect(
+                targetVal.expr.store(res.nonref_expr)
+            )
+            return context.constant(True)
+
+        return super().convert_to_type_with_target(context, instance, targetVal, conversionLevel, mayThrowOnFailure)
+
 
 class DictMakeIteratorWrapper(DictWrapperBase):
     def convert_method_call(self, context, expr, methodname, args, kwargs):
         if methodname == "__iter__" and not args and not kwargs:
             res = context.push(
                 # self.iteratorType is inherited from our specialized children
-                # who pick whether we're an interator over keys, values, items, etc.
+                # who pick whether we're an iterator over keys, values, items, etc.
                 self.iteratorType,
                 lambda instance:
                     instance.expr.ElementPtrIntegers(0, 0).store(-1)
+                    >> instance.expr.ElementPtrIntegers(0, 1).store(
+                        self.convert_len_native(expr)
+                    )
             )
 
             context.pushReference(
                 self,
-                res.expr.ElementPtrIntegers(0, 1)
+                res.expr.ElementPtrIntegers(0, 2)
             ).convert_copy_initialize(expr)
 
             return res
@@ -663,15 +657,35 @@ class DictIteratorWrapper(Wrapper):
 
     def getNativeLayoutType(self):
         return native_ast.Type.Struct(
-            element_types=(("pos", native_ast.Int64), ("dict", DictWrapper(self.dictType).getNativeLayoutType())),
+            element_types=(
+                ("pos", native_ast.Int64),
+                ("count", native_ast.Int64),
+                ("dict", DictWrapper(self.dictType).getNativeLayoutType())
+            ),
             name="const_dict_iterator"
         )
 
-    def convert_next(self, context, expr):
-        nextSlotIx = context.call_py_function(dict_next_slot, (self.refAs(context, expr, 1), self.refAs(context, expr, 0)), {})
+    def convert_fastnext(self, context, expr):
+        context.call_py_function(
+            checkDictSizeAndThrowIfChanged,
+            (
+                self.refAs(context, expr, 2),
+                self.refAs(context, expr, 1),
+            ),
+            {}
+        )
+
+        nextSlotIx = context.call_py_function(
+            table_next_slot,
+            (
+                self.refAs(context, expr, 2),
+                self.refAs(context, expr, 0)
+            ),
+            {}
+        )
 
         if nextSlotIx is None:
-            return None, None
+            return None
 
         context.pushEffect(
             expr.expr.ElementPtrIntegers(0, 0).store(
@@ -685,7 +699,7 @@ class DictIteratorWrapper(Wrapper):
 
         nextIx = context.pushReference(int, expr.expr.ElementPtrIntegers(0, 0))
 
-        return self.iteratedItemForReference(context, expr, nextIx), canContinue
+        return self.iteratedItemForReference(context, expr, nextIx).asPointerIf(canContinue)
 
     def refAs(self, context, expr, which):
         assert expr.expr_type == self
@@ -694,25 +708,28 @@ class DictIteratorWrapper(Wrapper):
             return context.pushReference(int, expr.expr.ElementPtrIntegers(0, 0))
 
         if which == 1:
+            return context.pushReference(int, expr.expr.ElementPtrIntegers(0, 1))
+
+        if which == 2:
             return context.pushReference(
                 self.dictType,
                 expr.expr
-                    .ElementPtrIntegers(0, 1)
+                    .ElementPtrIntegers(0, 2)
                     .cast(DictWrapper(self.dictType).getNativeLayoutType().pointer())
             )
 
     def convert_assign(self, context, expr, other):
         assert expr.isReference
 
-        for i in range(2):
+        for i in range(3):
             self.refAs(context, expr, i).convert_assign(self.refAs(context, other, i))
 
     def convert_copy_initialize(self, context, expr, other):
-        for i in range(2):
+        for i in range(3):
             self.refAs(context, expr, i).convert_copy_initialize(self.refAs(context, other, i))
 
     def convert_destroy(self, context, expr):
-        self.refAs(context, expr, 1).convert_destroy()
+        self.refAs(context, expr, 2).convert_destroy()
 
 
 class DictKeysIteratorWrapper(DictIteratorWrapper):
@@ -722,7 +739,7 @@ class DictKeysIteratorWrapper(DictIteratorWrapper):
     def iteratedItemForReference(self, context, expr, ixExpr):
         return DictWrapper(self.dictType).convert_method_call(
             context,
-            self.refAs(context, expr, 1),
+            self.refAs(context, expr, 2),
             "getKeyByIndexUnsafe",
             (ixExpr,),
             {}
@@ -736,7 +753,7 @@ class DictItemsIteratorWrapper(DictIteratorWrapper):
     def iteratedItemForReference(self, context, expr, ixExpr):
         return DictWrapper(self.dictType).convert_method_call(
             context,
-            self.refAs(context, expr, 1),
+            self.refAs(context, expr, 2),
             "getItemByIndexUnsafe",
             (ixExpr,),
             {}
@@ -750,7 +767,7 @@ class DictValuesIteratorWrapper(DictIteratorWrapper):
     def iteratedItemForReference(self, context, expr, ixExpr):
         return DictWrapper(self.dictType).convert_method_call(
             context,
-            self.refAs(context, expr, 1),
+            self.refAs(context, expr, 2),
             "getValueByIndexUnsafe",
             (ixExpr,),
             {}
