@@ -14,10 +14,11 @@
 
 from typed_python.compiler.typed_expression import TypedExpression
 from typed_python.compiler.type_wrappers.tuple_of_wrapper import TupleOrListOfWrapper
-from typed_python.compiler.type_wrappers.bound_compiled_method_wrapper import BoundCompiledMethodWrapper
+from typed_python.compiler.type_wrappers.bound_method_wrapper import BoundMethodWrapper
+from typed_python.compiler.conversion_level import ConversionLevel
 import typed_python.compiler.type_wrappers.runtime_functions as runtime_functions
 
-from typed_python import PointerTo
+from typed_python import ListOf
 
 import typed_python.compiler.native_ast as native_ast
 import typed_python.compiler
@@ -25,6 +26,96 @@ import typed_python.compiler
 typeWrapper = lambda t: typed_python.compiler.python_object_representation.typedPythonTypeToTypeWrapper(t)
 
 
+def list_of_extend(aList, extendWith):
+    if isinstance(extendWith, ListOf):
+        # don't iterate in case 'extendWith' is 'aList', in which case we could
+        # segfault if the list gets resized underneath us
+        for i in range(len(extendWith)):
+            aList.append(extendWith[i])
+    else:
+        for thing in extendWith:
+            aList.append(thing)
+
+
+def min(a, b):
+    return a if a < b else b
+
+
+def max(a, b):
+    return b if a < b else a
+
+
+def list_of_slice(aList, start, stop, step):
+    if step is None:
+        return list_of_slice(aList, start, stop, 1)
+    if step == 0:
+        raise ValueError("slice step cannot be zero")
+    if step > 0:
+        if start is None:
+            return list_of_slice(aList, 0, stop, step)
+
+        if stop is None:
+            return list_of_slice(aList, start, len(aList), step)
+
+        if start < 0:
+            start += len(aList)
+        if stop < 0:
+            stop += len(aList)
+
+        start = max(0, start)
+        stop = min(len(aList), stop)
+
+        outList = type(aList)()
+
+        count = max(0, (stop - start + (step - 1)) // step)
+
+        outList.reserve(count)
+        outList.setSizeUnsafe(count)
+
+        writeP = outList.pointerUnsafe(0)
+        readP = aList.pointerUnsafe(start)
+
+        i = start
+        while i < stop:
+            writeP.initialize(readP.get())
+            writeP += 1
+            i += step
+            readP += step
+
+        return outList
+    else:
+        if start is None:
+            return list_of_slice(aList, len(aList) - 1, stop, step)
+        if stop is None:
+            return list_of_slice(aList, start, -1 - len(aList), step)
+
+        if start < 0:
+            start += len(aList)
+        if stop < 0:
+            stop += len(aList)
+
+        start = min(len(aList) - 1, start)
+        stop = max(-1, stop)
+
+        outList = type(aList)()
+        count = max(0, (start - stop + (-step - 1)) // (-step))
+
+        outList.reserve(count)
+        outList.setSizeUnsafe(count)
+
+        writeP = outList.pointerUnsafe(0)
+        readP = aList.pointerUnsafe(start)
+
+        i = start
+        while i > stop:
+            writeP.initialize(readP.get())
+            writeP += 1
+            i += step
+            readP += step
+
+        return outList
+
+
 class ListOfWrapper(TupleOrListOfWrapper):
     is_pod = False
     is_empty = False
@@ -46,8 +137,9 @@ class ListOfWrapper(TupleOrListOfWrapper):
         return context.pushPod(int, expr.nonref_expr.ElementPtrIntegers(0, 3).load().cast(native_ast.Int64))
 
     def convert_attribute(self, context, instance, attr):
-        if attr in ("copy", "resize", "reserve", "reserved", "append", "clear", "pop", "setSizeUnsafe", "pointerUnsafe"):
-            return instance.changeType(BoundCompiledMethodWrapper(self, attr))
+        if attr in ("copy", "resize", "reserve", "reserved", "extend", "append",
+                    "clear", "pop", "setSizeUnsafe"):
+            return instance.changeType(BoundMethodWrapper.Make(self, attr))
 
         return super().convert_attribute(context, instance, attr)
 
@@ -83,29 +175,6 @@ class ListOfWrapper(TupleOrListOfWrapper):
                         native.call(instance, count)
                     )
 
-        if methodname == "pointerUnsafe":
-            if len(args) == 1:
-                count = args[0].toInt64()
-                if count is None:
-                    return
-
-                return context.pushPod(
-                    PointerTo(self.typeRepresentation.ElementType),
-                    instance.nonref_expr.ElementPtrIntegers(0, 4).load().cast(
-                        self.underlyingWrapperType.getNativeLayoutType().pointer()
-                    ).elemPtr(count.nonref_expr)
-                )
-
-        if methodname == "setSizeUnsafe":
-            if len(args) == 1:
-                count = args[0].toInt64()
-                if count is None:
-                    return
-
-                context.pushEffect(instance.nonref_expr.ElementPtrIntegers(0, 2).store(count.nonref_expr.cast(native_ast.Int32)))
-
-                return context.pushVoid()
-
         if methodname == "resize":
             if len(args) == 1:
                 count = args[0].toInt64()
@@ -127,7 +196,7 @@ class ListOfWrapper(TupleOrListOfWrapper):
                 if count is None:
                     return
 
-                val = args[1].convert_to_type(self.underlyingWrapperType)
+                val = args[1].convert_to_type(self.underlyingWrapperType, ConversionLevel.ImplicitContainers)
                 if val is None:
                     return
 
@@ -142,9 +211,13 @@ class ListOfWrapper(TupleOrListOfWrapper):
                     ).call(instance, count, val)
                 )
 
+        if methodname == "extend":
+            if len(args) == 1:
+                return context.call_py_function(list_of_extend, (instance, args[0]), {})
+
         if methodname == "append":
             if len(args) == 1:
-                val = args[0].convert_to_type(self.underlyingWrapperType)
+                val = args[0].convert_to_type(self.underlyingWrapperType, ConversionLevel.ImplicitContainers)
                 if val is None:
                     return
 
@@ -158,6 +231,7 @@ class ListOfWrapper(TupleOrListOfWrapper):
                         self.generateAppend
                     ).call(instance, val)
                 )
+
         if methodname == "copy":
             if len(args) == 0:
                 return context.push(
@@ -173,7 +247,16 @@ class ListOfWrapper(TupleOrListOfWrapper):
                 )
         if methodname == "clear":
             if len(args) == 0:
-                return self.convert_method_call(context, instance, "resize", (context.constant(0),), {})
+                return context.pushPod(
+                    None,
+                    context.converter.defineNativeFunction(
+                        'clear(' + self.typeRepresentation.__name__ + ")",
+                        ('util', self, 'clear'),
+                        [self],
+                        None,
+                        self.generateClear
+                    ).call(instance)
+                )
 
         if methodname == "reserve":
             if len(args) == 1:
@@ -221,7 +304,8 @@ class ListOfWrapper(TupleOrListOfWrapper):
         # we are just moving this - we assume no layouts have selfpointers throughout typed_python
         result = context.push(
             self.underlyingWrapperType,
-            lambda result: result.expr.store(inst.convert_getitem_unsafe(ix).nonref_expr)
+            lambda result: result.expr.store(inst.convert_getitem_unsafe(ix).nonref_expr),
+            wantsTeardown=False
         )
 
         context.pushEffect(
@@ -276,6 +360,14 @@ class ListOfWrapper(TupleOrListOfWrapper):
             listInst.nonref_expr.ElementPtrIntegers(0, 2).store(countInst.nonref_expr.cast(native_ast.Int32))
         )
 
+    def generateClear(self, context, out, listInst, arg=None):
+        with context.loop(listInst.convert_len()) as i:
+            listInst.convert_getitem_unsafe(i).convert_destroy()
+
+        context.pushEffect(
+            listInst.nonref_expr.ElementPtrIntegers(0, 2).store(native_ast.Int32.zero())
+        )
+
     def generateAppend(self, context, out, listInst, arg):
         with context.ifelse(listInst.convert_reserved() < listInst.convert_len()+1) as (if_needs_reserve, _):
             with if_needs_reserve:
@@ -293,7 +385,10 @@ class ListOfWrapper(TupleOrListOfWrapper):
         self.convert_method_call(context, out, "reserve", (listInst.convert_len(),), {})
 
         with context.loop(listInst.convert_len()) as i:
-            result = listInst.convert_getitem_unsafe(i).convert_to_type(typeWrapper(self.typeRepresentation.ElementType))
+            result = listInst.convert_getitem_unsafe(i).convert_to_type(
+                typeWrapper(self.typeRepresentation.ElementType),
+                ConversionLevel.Signature
+            )
             if result is None:
                 return None
             out.convert_getitem_unsafe(i).convert_copy_initialize(result)
@@ -348,11 +443,25 @@ class ListOfWrapper(TupleOrListOfWrapper):
             )  # data
         )
 
+    def convert_getslice(self, context, instance, lower, upper, step):
+        return context.call_py_function(
+            list_of_slice,
+            (
+                instance,
+                lower or context.constant(None),
+                upper or context.constant(None),
+                step or context.constant(None)
+            ),
+            {}
+        )
+
     def convert_setitem(self, context, expr, index, item):
         if item is None:
             return None
 
-        item = item.convert_to_type(self.underlyingWrapperType)
+        item = item.refToHeld()
+
+        item = item.convert_to_type(self.underlyingWrapperType, ConversionLevel.ImplicitContainers)
 
         if item is None:
             return None
@@ -384,6 +493,6 @@ class ListOfWrapper(TupleOrListOfWrapper):
             )
 
         if len(args) == 1 and not kwargs:
-            return args[0].convert_to_type(self, True)
+            return args[0].convert_to_type(self, ConversionLevel.New)
 
         return super().convert_type_call(context, typeInst, args, kwargs)
